"""BGP implementation for FAUCET."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress

from os_ken.lib import hub
from beka.beka import Beka
from faucet.valve_util import kill_on_exception


class BgpSpeakerKey:
    """Uniquely describe a BGP speaker."""

    def __init__(self, dp_id, vlan_vid, ipv):
        self.dp_id = dp_id
        self.vlan_vid = vlan_vid
        self.ipv = ipv

    def __str__(self):
        return "BGP speaker key DP ID: %u, VLAN VID: %u, IP version: %u" % (
            self.dp_id,
            self.vlan_vid,
            self.ipv,
        )

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()


class FaucetBgp:
    """Wrapper for Ryu BGP speaker."""

    exc_logname = None

    def __init__(self, logger, exc_logname, metrics, send_flow_msgs):
        self.logger = logger
        self.exc_logname = exc_logname
        self.metrics = metrics
        self._send_flow_msgs = send_flow_msgs
        self._dp_bgp_speakers = {}
        self._dp_bgp_rib = {}
        self._valves = None
        self.thread = None

    def _valve_vlan(self, dp_id, vlan_vid):
        valve = None
        vlan = None
        if dp_id in self._valves:
            if vlan_vid in self._valves[dp_id].dp.vlans:
                valve = self._valves[dp_id]
                vlan = valve.dp.vlans[vlan_vid]
        return (valve, vlan)

    @staticmethod
    def _neighbor_states(bgp_speaker):
        """Return state of each neighbor for a BGP speaker as a list."""
        neighbor_states = []
        if bgp_speaker is not None:
            neighbor_states = bgp_speaker.neighbor_states()
        return neighbor_states

    @kill_on_exception(exc_logname)
    def _bgp_up_handler(self, remote_ip, remote_as):
        self.logger.info("BGP peer router ID %s AS %s up" % (remote_ip, remote_as))

    @kill_on_exception(exc_logname)
    def _bgp_down_handler(self, remote_ip, remote_as):
        self.logger.info("BGP peer router ID %s AS %s down" % (remote_ip, remote_as))
        # TODO: delete RIB routes for down peer.

    @kill_on_exception(exc_logname)
    def _bgp_route_handler(self, path_change, bgp_speaker_key):
        """Handle a BGP change event.

        Args:
            path_change (ryu.services.protocols.bgp.bgpspeaker.EventPrefix): path change
        """
        dp_id = bgp_speaker_key.dp_id
        vlan_vid = bgp_speaker_key.vlan_vid
        valve, vlan = self._valve_vlan(dp_id, vlan_vid)
        if vlan is None:
            return
        prefix = ipaddress.ip_network(str(path_change.prefix))
        route_str = "BGP route %s" % prefix

        if path_change.next_hop:
            nexthop = ipaddress.ip_address(str(path_change.next_hop))
            route_str = "BGP route %s nexthop %s" % (prefix, nexthop)

            if vlan.is_faucet_vip(nexthop):
                self.logger.error(
                    "Skipping %s because nexthop cannot be us" % route_str
                )
                return

            if valve.router_vlan_for_ip_gw(vlan, nexthop) is None:
                self.logger.info(
                    "Skipping %s because nexthop not in %s" % (route_str, vlan)
                )
                return

        if bgp_speaker_key not in self._dp_bgp_rib:
            self._dp_bgp_rib[bgp_speaker_key] = {}

        flowmods = []
        if path_change.is_withdraw:
            self.logger.info("withdraw %s", route_str)
            if prefix in self._dp_bgp_rib[bgp_speaker_key]:
                del self._dp_bgp_rib[bgp_speaker_key][prefix]
            flowmods = valve.del_route(vlan, prefix)
        else:
            self.logger.info("add %s", route_str)
            self._dp_bgp_rib[bgp_speaker_key][prefix] = nexthop
            flowmods = valve.add_route(vlan, nexthop, prefix)
        if flowmods:
            self._send_flow_msgs(valve, flowmods)

    @staticmethod
    def _vlan_prefixes_by_ipv(vlan, ipv):
        vlan_prefixes = [
            (str(faucet_vip), str(faucet_vip.ip))
            for faucet_vip in vlan.faucet_vips_by_ipv(ipv)
        ]
        vlan_prefixes.extend(
            [
                (str(ip_dst), str(ip_gw))
                for ip_dst, ip_gw in vlan.routes_by_ipv(ipv).items()
            ]
        )
        return vlan_prefixes

    def _create_bgp_speaker_for_vlan(self, bgp_speaker_key, bgp_router):
        """Set up BGP speaker for an individual VLAN if required.

        Args:
            bgp_speaker_key (BgpSpeakerKey): BGP speaker key.
            bgp_router: Router.
        Returns:
            ryu.services.protocols.bgp.bgpspeaker.BGPSpeaker: BGP speaker.
        """
        server_address = sorted(
            bgp_router.bgp_server_addresses_by_ipv(bgp_speaker_key.ipv)
        )[0]
        beka = Beka(
            local_address=str(server_address),
            bgp_port=bgp_router.bgp_port(),
            local_as=bgp_router.bgp_as(),
            router_id=bgp_router.bgp_routerid(),
            peer_up_handler=self._bgp_up_handler,
            peer_down_handler=self._bgp_down_handler,
            route_handler=lambda x: self._bgp_route_handler(x, bgp_speaker_key),
            error_handler=self.logger.warning,
        )
        for ip_dst, ip_gw in self._vlan_prefixes_by_ipv(
            bgp_router.bgp_vlan(), bgp_speaker_key.ipv
        ):
            beka.add_route(prefix=str(ip_dst), next_hop=str(ip_gw))
        for bgp_neighbor_address in bgp_router.bgp_neighbor_addresses_by_ipv(
            bgp_speaker_key.ipv
        ):
            beka.add_neighbor(
                connect_mode=bgp_router.bgp_connect_mode(),
                peer_ip=str(bgp_neighbor_address),
                peer_as=bgp_router.bgp_neighbor_as(),
            )
        self.thread = hub.spawn(beka.run)
        self.thread.name = "beka"
        return beka

    def shutdown_bgp_speakers(self):
        """Shutdown any active BGP speakers."""
        for bgp_speaker in self._dp_bgp_speakers.values():
            bgp_speaker.shutdown()
        self._dp_bgp_speakers = {}

    def _add_bgp_speaker(self, valve, bgp_speaker_key, bgp_router):
        if bgp_speaker_key in self._dp_bgp_speakers:
            self.logger.info(
                "Skipping re/configuration of existing %s" % bgp_speaker_key
            )
            bgp_speaker = self._dp_bgp_speakers[bgp_speaker_key]
            if bgp_speaker_key in self._dp_bgp_rib:
                # Re-add routes (to avoid flapping BGP even when VLAN cold starts).
                for prefix, nexthop in self._dp_bgp_rib[bgp_speaker_key].items():
                    self.logger.info("Re-adding %s via %s" % (prefix, nexthop))
                    bgp_vlan = bgp_router.bgp_vlan()
                    flowmods = valve.add_route(bgp_vlan, nexthop, prefix)
                    if flowmods:
                        self._send_flow_msgs(valve, flowmods)
        else:
            self.logger.info("Adding %s" % bgp_speaker_key)
            bgp_speaker = self._create_bgp_speaker_for_vlan(bgp_speaker_key, bgp_router)
        return {bgp_speaker_key: bgp_speaker}

    def _add_valve_bgp_speakers(self, valve):
        bgp_speakers = {}
        bgp_routers = valve.dp.bgp_routers()
        if bgp_routers:
            dp_id = valve.dp.dp_id
            for bgp_router in bgp_routers:
                bgp_vlan = bgp_router.bgp_vlan()
                vlan_vid = bgp_vlan.vid
                for ipv in bgp_router.bgp_ipvs():
                    bgp_speaker_key = BgpSpeakerKey(dp_id, vlan_vid, ipv)
                    bgp_speakers.update(
                        self._add_bgp_speaker(valve, bgp_speaker_key, bgp_router)
                    )
        return bgp_speakers

    def reset(self, valves):
        """Set up a BGP speaker for every VLAN that requires it."""
        # TODO: port status changes should cause us to withdraw a route.
        new_dp_bgp_speakers = {}
        if valves:
            for valve in valves.values():
                new_dp_bgp_speakers.update(self._add_valve_bgp_speakers(valve))
        # TODO: shutdown and remove deconfigured BGP speakers.
        for bgp_speaker_key, old_bgp_speaker in self._dp_bgp_speakers.items():
            if bgp_speaker_key not in new_dp_bgp_speakers:
                new_dp_bgp_speakers[bgp_speaker_key] = old_bgp_speaker
        self._dp_bgp_speakers = new_dp_bgp_speakers
        self._valves = valves

    def update_metrics(self, _now):
        """Update BGP metrics."""
        for bgp_speaker_key, bgp_speaker in self._dp_bgp_speakers.items():
            dp_id = bgp_speaker_key.dp_id
            vlan_vid = bgp_speaker_key.vlan_vid
            ipv = bgp_speaker_key.ipv
            valve, vlan = self._valve_vlan(dp_id, vlan_vid)
            if vlan is None:
                continue
            neighbor_states = self._neighbor_states(bgp_speaker)
            for neighbor, neighbor_state in neighbor_states:
                neighbor_labels = dict(
                    valve.dp.base_prom_labels(), vlan=vlan.vid, neighbor=neighbor
                )
                self.metrics.bgp_neighbor_uptime_seconds.labels(  # pylint: disable=no-member
                    **neighbor_labels
                ).set(
                    neighbor_state["info"]["uptime"]
                )
                self.metrics.bgp_neighbor_routes.labels(  # pylint: disable=no-member
                    **dict(neighbor_labels, ipv=ipv)
                ).set(vlan.route_count_by_ipv(ipv))



"""Unit tests for VLAN"""

import unittest
from ipaddress import ip_address, ip_network, ip_interface

from faucet.vlan import VLAN


class FaucetVLANMethodTest(unittest.TestCase):
    """Initialises VLANs with different configs and sanity checks the associated methods"""

    def test_ipvs_no_ips(self):
        """Tests the ipvs() method with no vips"""

        vlan = VLAN(1, 1, {})
        self.assertEqual(len(vlan.ipvs()), 0)

    def test_ipvs_ipv4(self):
        """Tests the ipvs() method with an IPv4 vip"""

        vlan_config = {"faucet_vips": ["10.0.0.254/24"]}

        vlan = VLAN(1, 1, vlan_config)
        self.assertIn(4, vlan.ipvs())
        self.assertNotIn(6, vlan.ipvs())

    def test_ipvs_ipv6(self):
        """Tests the ipvs() method with an IPv6 vip"""

        vlan_config = {"faucet_vips": ["2001::1/16"]}

        vlan = VLAN(1, 1, vlan_config)
        self.assertIn(6, vlan.ipvs())
        self.assertNotIn(4, vlan.ipvs())

    def test_ipvs_ipv4_ipv6(self):
        """Tests the ipvs() method with both IPv4 and IPv6 vips"""

        vlan_config = {"faucet_vips": ["2001::1/16", "fe80::1/64", "10.0.0.254/24"]}

        vlan = VLAN(1, 1, vlan_config)
        self.assertIn(4, vlan.ipvs())
        self.assertIn(6, vlan.ipvs())

    def test_faucet_vips_by_ipv_none(self):
        """Tests the faucet_vips_by_ipv() method when there are no vips"""

        vlan = VLAN(1, 1, {})
        self.assertEqual(len(vlan.faucet_vips_by_ipv(4)), 0)
        self.assertEqual(len(vlan.faucet_vips_by_ipv(6)), 0)

    def test_faucet_vips_by_ipv_both(self):
        """Tests the faucet_vips_by_ipv() method when there are both IPv4 and IPv6 vips"""

        vlan_config = {"faucet_vips": ["2001::1/16", "fe80::1/64", "10.0.0.254/24"]}

        vlan = VLAN(1, 1, vlan_config)
        self.assertEqual(
            set(vlan.faucet_vips_by_ipv(4)), set([ip_interface("10.0.0.254/24")])
        )
        self.assertEqual(
            set(vlan.faucet_vips_by_ipv(6)),
            set([ip_interface("2001::1/16"), ip_interface("fe80::1/64")]),
        )

    def test_routes_by_ipv_none(self):
        """Tests the routes_by_ipv() and route_count_by_ipv() methods with no routes"""

        vlan = VLAN(1, 1, {})
        self.assertEqual(vlan.routes_by_ipv(4), {})
        self.assertEqual(vlan.routes_by_ipv(6), {})
        self.assertEqual(vlan.route_count_by_ipv(4), 0)
        self.assertEqual(vlan.route_count_by_ipv(6), 0)

    def test_routes_by_ipv_both(self):
        """Tests the routes_by_ipv() and route_count_by_ipv() methods with both
        IPv4 and IPv6 routes"""

        vlan_config = {
            "routes": [
                {"route": {"ip_dst": "10.99.99.0/24", "ip_gw": "10.0.0.1"}},
                {"route": {"ip_dst": "10.99.98.0/24", "ip_gw": "10.0.0.99"}},
                {"route": {"ip_dst": "10.99.97.0/24", "ip_gw": "10.0.0.99"}},
                {"route": {"ip_dst": "fc00::10:0/112", "ip_gw": "fc00::1:1"}},
                {"route": {"ip_dst": "fc00::20:0/112", "ip_gw": "fc00::1:99"}},
            ],
        }

        vlan = VLAN(1, 1, vlan_config)

        self.assertEqual(
            vlan.routes_by_ipv(4),
            {
                ip_network("10.99.99.0/24"): ip_address("10.0.0.1"),
                ip_network("10.99.98.0/24"): ip_address("10.0.0.99"),
                ip_network("10.99.97.0/24"): ip_address("10.0.0.99"),
            },
        )
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {
                ip_network("fc00::10:0/112"): ip_address("fc00::1:1"),
                ip_network("fc00::20:0/112"): ip_address("fc00::1:99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(4), 3)
        self.assertEqual(vlan.route_count_by_ipv(6), 2)

    def test_modify_routes_v4(self):
        """Tests the add_route() and remove_route() methods with IPv4 routes"""

        vlan = VLAN(1, 1, {})

        self.assertEqual(vlan.routes_by_ipv(4), {})
        vlan.add_route(ip_network("10.99.99.0/24"), ip_address("10.0.0.1"))
        vlan.add_route(ip_network("10.99.98.0/24"), ip_address("10.0.0.99"))
        self.assertEqual(
            vlan.routes_by_ipv(4),
            {
                ip_network("10.99.99.0/24"): ip_address("10.0.0.1"),
                ip_network("10.99.98.0/24"): ip_address("10.0.0.99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(4), 2)
        vlan.del_route(ip_network("10.99.99.0/24"))
        self.assertEqual(
            vlan.routes_by_ipv(4),
            {ip_network("10.99.98.0/24"): ip_address("10.0.0.99")},
        )
        self.assertEqual(vlan.route_count_by_ipv(4), 1)
        vlan.del_route(ip_network("10.99.98.0/24"))
        self.assertEqual(vlan.route_count_by_ipv(4), 0)
        self.assertEqual(vlan.routes_by_ipv(4), {})

    def test_modify_routes_v6(self):
        """Tests the add_route() and remove_route() methods with IPv4 routes"""

        vlan = VLAN(1, 1, {})

        self.assertEqual(vlan.routes_by_ipv(6), {})
        vlan.add_route(ip_network("fc00::10:0/112"), ip_address("fc00::1:1"))
        vlan.add_route(ip_network("fc00::20:0/112"), ip_address("fc00::1:99"))
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {
                ip_network("fc00::10:0/112"): ip_address("fc00::1:1"),
                ip_network("fc00::20:0/112"): ip_address("fc00::1:99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(6), 2)
        vlan.del_route(ip_network("fc00::10:0/112"))
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {ip_network("fc00::20:0/112"): ip_address("fc00::1:99")},
        )
        self.assertEqual(vlan.route_count_by_ipv(6), 1)
        vlan.del_route(ip_network("fc00::20:0/112"))
        self.assertEqual(vlan.route_count_by_ipv(6), 0)
        self.assertEqual(vlan.routes_by_ipv(6), {})

    def test_modify_routes_static_v4(self):
        """Tests the add_route() and remove_route() methods,
        starting with configured static routes for IPv4"""

        vlan_config = {
            "routes": [
                {"route": {"ip_dst": "10.99.97.0/24", "ip_gw": "10.0.0.99"}},
            ],
        }

        vlan = VLAN(1, 1, vlan_config)

        self.assertEqual(
            vlan.routes_by_ipv(4),
            {ip_network("10.99.97.0/24"): ip_address("10.0.0.99")},
        )
        vlan.add_route(ip_network("10.99.99.0/24"), ip_address("10.0.0.1"))
        vlan.add_route(ip_network("10.99.98.0/24"), ip_address("10.0.0.99"))
        self.assertEqual(
            vlan.routes_by_ipv(4),
            {
                ip_network("10.99.99.0/24"): ip_address("10.0.0.1"),
                ip_network("10.99.98.0/24"): ip_address("10.0.0.99"),
                ip_network("10.99.97.0/24"): ip_address("10.0.0.99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(4), 3)
        vlan.del_route(ip_network("10.99.99.0/24"))
        self.assertEqual(
            vlan.routes_by_ipv(4),
            {
                ip_network("10.99.97.0/24"): ip_address("10.0.0.99"),
                ip_network("10.99.98.0/24"): ip_address("10.0.0.99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(4), 2)
        vlan.del_route(ip_network("10.99.98.0/24"))
        self.assertEqual(vlan.route_count_by_ipv(4), 1)
        self.assertEqual(
            vlan.routes_by_ipv(4),
            {ip_network("10.99.97.0/24"): ip_address("10.0.0.99")},
        )

    def test_modify_routes_static_v6(self):
        """Tests the add_route() and remove_route() methods,
        starting with configured static routes for IPv6"""

        vlan_config = {
            "routes": [
                {"route": {"ip_dst": "fc00::30:0/112", "ip_gw": "fc00::1:99"}},
            ],
        }

        vlan = VLAN(1, 1, vlan_config)

        self.assertEqual(
            vlan.routes_by_ipv(6),
            {ip_network("fc00::30:0/112"): ip_address("fc00::1:99")},
        )
        vlan.add_route(ip_network("fc00::10:0/112"), ip_address("fc00::1:1"))
        vlan.add_route(ip_network("fc00::20:0/112"), ip_address("fc00::1:99"))
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {
                ip_network("fc00::10:0/112"): ip_address("fc00::1:1"),
                ip_network("fc00::20:0/112"): ip_address("fc00::1:99"),
                ip_network("fc00::30:0/112"): ip_address("fc00::1:99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(6), 3)
        vlan.del_route(ip_network("fc00::10:0/112"))
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {
                ip_network("fc00::30:0/112"): ip_address("fc00::1:99"),
                ip_network("fc00::20:0/112"): ip_address("fc00::1:99"),
            },
        )
        self.assertEqual(vlan.route_count_by_ipv(6), 2)
        vlan.del_route(ip_network("fc00::20:0/112"))
        self.assertEqual(vlan.route_count_by_ipv(6), 1)
        self.assertEqual(
            vlan.routes_by_ipv(6),
            {ip_network("fc00::30:0/112"): ip_address("fc00::1:99")},
        )


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""Mininet Topo class with YAML config generator"""

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import string

from mininet.log import output
from mininet.topo import Topo

from clib import mininet_test_util
from clib.mininet_test_topo import (
    FaucetHost,
    VLANHost,
    FaucetSwitch,
    NoControllerFaucetSwitch,
)
from clib.valve_test_lib import yaml_dump


class GenerationError(Exception):
    """Indicates a problem with generating the configuration file"""


class FaucetTopoGenerator(Topo):
    """Creates a mininet topology and then provides a method to generate a YAML config file"""

    # Host CPU option
    CPUF = 0.5
    # Link delay option
    DELAY = "1ms"

    # Switch index map to switch name
    switches_by_id = {}
    # Switch index map to switch dpid
    dpids_by_id = {}
    # Host index map to host name
    hosts_by_id = {}

    # Generated hardware switch name
    hw_name = None
    # DPID of the hardware switch
    hw_dpid = None
    # List of port order for the hardware switch
    hw_ports = None

    # Function to resolve serial numbers
    get_serialno = None

    # Additional mininet host options
    host_options = None

    # The generated starting port for each switch
    start_port = None
    # The port order for each switch
    port_order = None

    def get_dpids(self):
        """Returns list of DPIDs in switch index order"""
        return [self.dpids_by_id[key] for key in sorted(self.dpids_by_id)]

    def _create_link_port_map(self):
        """Switch pair link map to list of ports for that pair"""
        port_maps = {}
        for i, name in self.switches_by_id.items():
            for port, link in self.ports[name].items():
                if self.isSwitch(link[0]):
                    peer_id = self.nodeInfo(link[0])["switch_n"]
                    port_maps.setdefault((i, peer_id), [])
                    port_maps[(i, peer_id)].append(port)
        return port_maps

    def _create_host_port_map(self):
        """Host map to linked switches to list of ports from switch to host"""
        host_port_map = {}
        for host, name in self.hosts_by_id.items():
            host_port_map.setdefault(host, {})
            for link in self.ports[name].values():
                switch_id = self.nodeInfo(link[0])["switch_n"]
                host_port_map[host].setdefault(switch_id, [])
                host_port_map[host][switch_id].append(link[1])
        return host_port_map

    def _create_port_map(self):
        """Create a map from port to the true port"""
        port_maps = {}
        for i, dpid in self.dpids_by_id.items():
            switch_name = self.switches_by_id[i]
            ports = self.ports[switch_name].keys()
            port_maps[dpid] = {"port_%d" % i: port for i, port in enumerate(ports)}
        return port_maps

    def create_port_maps(self):
        """Return host port maps and link port maps"""
        return (
            self._create_port_map(),
            self._create_host_port_map(),
            self._create_link_port_map(),
        )

    def rand_sequential_dpid(self, i):
        """Returns a DPID that is sequential to the other DPIDs in the topology"""
        # Some functions in the Faucet code relies on the DPID for determining the order
        #   of an operation; E.g. stack root nominations. This could mean that by using randomized
        #   DPIDs, the test could be flaky due to the random ordering of the DPIDs
        if i == 0 and self.hw_dpid:
            # HW_DPID probably won't obey the starting DPID
            return self.hw_dpid
        if not self.dpids_by_id or (i == 1 and self.hw_dpid):
            # Choose a random starting DPID if none exist
            # If the hw_dpid exists, then switch0 will have it,
            #   in this case it might ruin the sequential dpid choosing
            #   so just randomly choose a DPID
            reserved_range = 100
            min_dpid = 2 + self.num_dps
            max_dpid = 2**32 - reserved_range - self.num_dps - 1
            while True:
                dpid = random.randint(min_dpid, max_dpid) + reserved_range
                if dpid not in self.dpids_by_id.values():
                    return str(dpid)
        # Get most recent DPID and increase by 1
        prev_dpid = sorted(self.dpids_by_id.values(), reverse=self.descending_dpids)[-1]
        while True:
            if self.descending_dpids:
                dpid = int(prev_dpid) - 1
            else:
                dpid = int(prev_dpid) + 1
            if dpid not in self.dpids_by_id.values():
                return str(dpid)

    @staticmethod
    def vlan_name(i):
        """VLAN name"""
        return "vlan-%i" % (i + 1)

    @staticmethod
    def vlan_vid(i):
        """VLAN VID value"""
        return (i + 1) * 100

    @staticmethod
    def router_name(i):
        """Router name"""
        return "router-%s" % (i + 1)

    def __init__(self, *args, **kwargs):
        self.switches_by_id = {}
        self.dpids_by_id = {}
        self.hosts_by_id = {}
        self.num_dps = None
        self.descending_dpids = None
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_sid_prefix(ports_served):
        """Return a unique switch/host prefix for a test."""
        # Linux tools require short interface names.
        id_chars = "".join(
            sorted(string.ascii_letters + string.digits)  # pytype: disable=module-attr
        )
        id_a = int(ports_served / len(id_chars))
        id_b = ports_served - (id_a * len(id_chars))
        return "%s%s" % (id_chars[id_a], id_chars[id_b])

    @staticmethod
    def extend_port_order(port_order=None, max_length=16):
        """
        Extends the pattern of port_port order up to max_length

        Args:
            port_order (list): List of integers in an order to extend
            max_length (int): Maximum length to extend the list to
        """
        if not port_order:
            return list(range(max_length + 1))
        if len(port_order) >= max_length:
            return port_order
        extend_order = []
        order = port_order
        start_port = max(port_order) + 1
        while len(port_order) + len(extend_order) < max_length:
            for i in order:
                extend_order.append(start_port + i)
                if len(port_order) + len(extend_order) >= max_length:
                    break
            start_port = max(extend_order) + 1
        return port_order + extend_order

    def _generate_sid_prefix(self):
        """Returns a sid prefix for a node in the topology"""
        return self._get_sid_prefix(self.get_serialno(self.ports_sock, self.test_name))

    def _create_next_port(self, switch_name):
        """
        Creates and returns the next port number for a switch

        Args:
            switch_name (str): The name of the switch to generate the next port
        """
        index = 0
        if switch_name in self.ports:
            index = len(self.ports[switch_name])
        if self.hw_name and switch_name == self.hw_name and self.hw_ports:
            return self.hw_ports[self.port_order[index]]
        return self.start_port + self.port_order[index]

    def _add_host(self, host_index, vlans):
        """
        Adds a untagged/tagged host to the topology

        Args:
            sid_prefix (str): SID prefix to generate the host name
            host_index (int): Host index to generate the host name
            vlans (list/None/int): Type of host/vlans the host belongs to
        """
        sid_prefix = self._generate_sid_prefix()
        host_opts = self.host_options.get(host_index, {})
        host_name = None
        if "cls" in host_opts:
            host_name = "e%s%1.1u" % (sid_prefix, host_index + 1)
        else:
            if isinstance(vlans, int) or vlans is None:
                host_name = "u%s%1.1u" % (sid_prefix, host_index + 1)
                host_opts["cls"] = FaucetHost
            elif isinstance(vlans, list):
                host_name = "t%s%1.1u" % (sid_prefix, host_index + 1)
                host_opts["vlans"] = [
                    FaucetTopoGenerator.vlan_vid(vlan) for vlan in vlans
                ]
                host_opts["cls"] = VLANHost
            else:
                raise GenerationError("Unknown host type")
        self.hosts_by_id[host_index] = host_name
        return self.addHost(
            cpu=self.CPUF,
            host_n=host_index,
            name=host_name,
            config_vlans=vlans,
            **host_opts
        )

    def _add_faucet_switch(self, switch_index):
        """
        Adds a Faucet switch to the topology

        Args:
            sid_prefix (str): SID prefix to generate the switch name
            switch_index (int): Switch index to generate the host name
            dpid (int): Switch DP ID
        """
        sid_prefix = self._generate_sid_prefix()
        switch_cls = FaucetSwitch
        switch_name = "s%s" % sid_prefix
        if switch_index == 0 and self.hw_dpid:
            self.hw_name = switch_name
            self.dpids_by_id[switch_index] = self.hw_dpid
            dpid = str(int(self.hw_dpid) + 1)
            output(
                "bridging hardware switch DPID %s (%x) dataplane via OVS DPID %s (%x)\n"
                % (self.hw_dpid, int(self.hw_dpid), dpid, int(dpid))
            )
            switch_cls = NoControllerFaucetSwitch
        else:
            dpid = self.rand_sequential_dpid(switch_index)
            self.dpids_by_id[switch_index] = dpid
        self.switches_by_id[switch_index] = switch_name
        return self.addSwitch(
            name=switch_name,
            cls=switch_cls,
            datapath=self.ovs_type,
            dpid=mininet_test_util.mininet_dpid(dpid),
            switch_n=switch_index,
        )

    def _add_link(self, node, peer_node, vlans):
        """
        Creates and adds a link between two nodes to the topology

        Args:
            node (str): Name of the node for the link, NOTE: should ALWAYS be a switch
            peer_node (str): Name of the peer node for the link
            vlans (list/None/int): Type of the link
        """
        port1, port2 = None, None
        opts = {}
        if self.isSwitch(node):
            # Node is a switch, create port
            port1 = self._create_next_port(node)
        if self.isSwitch(peer_node):
            # Peer node is a switch, create port
            port2 = self._create_next_port(peer_node)
        else:
            # Peer node is a host, use delay & htb options
            opts["delay"] = self.DELAY
            opts["use_htb"] = True
        return self.addLink(
            node, peer_node, port1=port1, port2=port2, **opts, config_vlans=vlans
        )

    def add_switch_topology(self, switch_links, link_vlans):
        """
        Adds the switches and switch-switch links to the network topology
        Tagged links are mapped to a list of vlan indices whereas untagged links
            are mapped to a single vlan index, stack links are mapped to None

        Args:
            switch_topology (list): List of link tuples of switch indices (u, v)
            link_vlans (dict): Link tuple of switch indices (u, v) mapping to vlans
        """
        for u_id, v_id in switch_links:
            if u_id not in self.switches_by_id:
                self._add_faucet_switch(u_id)
            if v_id not in self.switches_by_id:
                self._add_faucet_switch(v_id)
            u_name = self.switches_by_id[u_id]
            v_name = self.switches_by_id[v_id]
            self._add_link(u_name, v_name, link_vlans[(u_id, v_id)])

    def add_host_topology(self, host_links, host_vlans):
        """
        Adds the hosts and host-switch links to the network topology
        Tagged hosts are mapped to a list of vlan indices whereas untagged hosts
            are mapped to a single vlan index

        Args:
            host_links (dict): Host index key to list of dp indices
            host_vlans (dict): Host index key to vlan index/indices
        """
        for h_id, links in host_links.items():
            vlans = host_vlans[h_id]
            if h_id not in self.hosts_by_id:
                self._add_host(h_id, vlans)
            host_name = self.hosts_by_id[h_id]
            for dp_i in links:
                if dp_i not in self.switches_by_id:
                    self._add_faucet_switch(dp_i)
                switch_name = self.switches_by_id[dp_i]
                self._add_link(switch_name, host_name, vlans)

    # pylint: disable=arguments-differ
    def build(
        self,
        ovs_type,
        ports_sock,
        test_name,
        num_dps,
        descending_dpids,
        host_links,
        host_vlans,
        switch_links,
        link_vlans,
        hw_dpid=None,
        hw_ports=None,
        port_order=None,
        start_port=5,
        get_serialno=mininet_test_util.get_serialno,
        host_options=None,
    ):
        """
        Creates a Faucet mininet topology

        Args:
            ovs_type (str): The OVS switch type
            ports_sock (str): Port socket
            test_name (str): Name of the test creating the mininet topology
            host_links (dict): Host index key to list of dp indices
            host_vlans (dict): Host index key to vlan index/indices
            switch_links (list): List of link tuples of switch indices (u, v)
            link_vlans (dict): Link tuple of switch indices (u, v) mapping to vlans
            hw_dpid (int): DP ID of the hardware switch to connect to the topology
            hw_ports (list): Map of the OVS bridge port index to hardware port number
            port_order (list): List of integers in order for a switch port index order
            start_port (int): The minimum start port number for all switch port numbers
            get_serialno (func): Function to get the serial no.
            host_options (dict): Host index map to additional mininet host options
        """
        # Additional test generation information
        self.ovs_type = ovs_type
        self.ports_sock = ports_sock
        self.test_name = test_name
        self.get_serialno = get_serialno
        self.num_dps = num_dps
        self.descending_dpids = descending_dpids

        # Information for hardware switches
        self.hw_dpid = hw_dpid
        self.hw_ports = sorted(hw_ports) if hw_ports else []

        # Additional information for special hosts
        self.host_options = host_options if host_options else {}

        # Generate a port order for all of the switches to use
        max_ports = (len(switch_links) * 2) + len(host_links)
        self.start_port = start_port
        self.port_order = self.extend_port_order(port_order, max_length=max_ports)

        # Build the network topology
        self.add_switch_topology(switch_links, link_vlans)
        self.add_host_topology(host_links, host_vlans)

    @staticmethod
    def get_acls_config(acl_options):
        """Return the ACLs in dictionary format for the configuration file"""
        return acl_options.copy()

    def get_dps_config(self, dp_options, host_options, link_options, ignored_switches):
        """Return the DPs in dictionary format for the configuration file"""
        dps_config = {}

        def get_interface_config(
            link_name, src_port, dst_node, dst_port, vlans, options, ignored
        ):
            interface_config = {}
            _type = "switch-switch" if dst_port else "switch-host"
            if ignored:
                # Link is to an outside network, so treat it as a output only link with more
                #   specific options defined in the options dictionary
                interface_config = {
                    "name": "b%u" % src_port,
                    "description": "output only %s" % link_name,
                }
            elif isinstance(vlans, int):
                # Untagged link
                interface_config = {
                    "name": "b%u" % src_port,
                    "description": "untagged %s" % link_name,
                    "native_vlan": self.vlan_name(vlans),
                }
            elif isinstance(vlans, list):
                # Tagged link
                interface_config = {
                    "name": "b%u" % src_port,
                    "description": "tagged %s" % link_name,
                    "tagged_vlans": [self.vlan_name(vlan) for vlan in vlans],
                }
            elif dst_node and dst_port:
                # Stack link
                interface_config = {
                    "name": "b%u" % src_port,
                    "description": "stack %s" % link_name,
                    "stack": {"dp": dst_node, "port": dst_port},
                }
            elif vlans is None:
                # output only link or coprocessor, leave to more specific options to handle
                interface_config = {
                    "name": "b%u" % src_port,
                    "description": "output only %s" % link_name,
                }
            else:
                raise GenerationError("Unknown %s link type %s" % (_type, vlans))
            if options:
                for option_key, option_value in options.items():
                    interface_config[option_key] = option_value
            return interface_config

        def add_dp_config(src_node, dst_node, link_key, link_info, reverse=False):
            dp_config = dps_config[src_node]
            src_info, dst_info = self.nodeInfo(src_node), self.nodeInfo(dst_node)
            vlans = link_info["config_vlans"]
            src_id = src_info["switch_n"]
            dp_config.setdefault("interfaces", {})
            options = {}
            ignored = False
            if self.isSwitch(dst_node):
                # Generate switch-switch config link
                if reverse:
                    src_port, dst_port = link_info["port2"], link_info["port1"]
                else:
                    src_port, dst_port = link_info["port1"], link_info["port2"]
                link_name = "link #%s to %s:%s" % (link_key, dst_node, dst_port)
                options = {}
                dst_id = dst_info["switch_n"]
                if link_options and (src_id, dst_id) in link_options:
                    options.update(link_options[(src_id, dst_id)])
                if dst_id in ignored_switches:
                    ignored = True
            else:
                # Generate host-switch config link
                src_port, dst_port = link_info["port1"], None
                link_name = "link #%s to %s" % (link_key, dst_node)
                host_n = dst_info["host_n"]
                if host_options and host_n in host_options:
                    options = host_options[host_n]
            dp_config["interfaces"].setdefault(  # pytype: disable=attribute-error
                src_port,
                get_interface_config(
                    link_name, src_port, dst_node, dst_port, vlans, options, ignored
                ),
            )

        for links in self.links(withKeys=True, withInfo=True):
            src_node, dst_node, link_key, link_info = links
            src_info = self.nodeInfo(src_node)
            dst_info = self.nodeInfo(dst_node)
            if self.isSwitch(src_node) and src_info["switch_n"] not in ignored_switches:
                dps_config.setdefault(src_node, {})
                src_dpid = self.dpids_by_id[src_info["switch_n"]]
                dps_config[src_node].setdefault("dp_id", int(src_dpid))
                add_dp_config(src_node, dst_node, link_key, link_info)
            if self.isSwitch(dst_node) and dst_info["switch_n"] not in ignored_switches:
                dps_config.setdefault(dst_node, {})
                dst_dpid = self.dpids_by_id[dst_info["switch_n"]]
                dps_config[dst_node].setdefault("dp_id", int(dst_dpid))
                add_dp_config(dst_node, src_node, link_key, link_info, True)
        if dp_options:
            for dp_i, options in dp_options.items():
                switch_name = self.switches_by_id[dp_i]
                dps_config.setdefault(switch_name, {})
                for option_key, option_value in options.items():
                    dps_config[switch_name][option_key] = option_value
        return dps_config

    def get_vlans_config(self, n_vlans, vlan_options):
        """
        Return the VLANs in dictionary format for the YAML configuration file

        Args:
            n_vlans (int): Number of VLANs to generate
            vlan_options (dict): Additional options for each VLAN, keyed by vlan index
        """
        vlans_config = {}
        for vlan in range(n_vlans):
            vlan_name = self.vlan_name(vlan)
            vlans_config[vlan_name] = {"vid": FaucetTopoGenerator.vlan_vid(vlan)}
        if vlan_options:
            for vlan, options in vlan_options.items():
                vlan_name = self.vlan_name(vlan)
                for option_key, option_value in options.items():
                    vlans_config[vlan_name][option_key] = option_value
        return vlans_config

    def get_routers_config(self, routers, router_options):
        """
        Return the routers in dictionary format for the configuration file

        Args:
            routers (dict): Router index to list of VLANs in the router
            router_options (dict): Additional options for each router, keyed by router index
        """
        routers_config = {}
        for router, vlans in routers.items():
            routers_config[self.router_name(router)] = {
                "vlans": [self.vlan_name(vlan) for vlan in vlans]
            }
        if router_options:
            for router, options in router_options.items():
                router_name = self.router_name(router)
                for option_key, option_value in options.items():
                    routers_config[router_name][option_key] = option_value
        return routers_config

    def get_config(
        self,
        n_vlans,
        acl_options=None,
        dp_options=None,
        host_options=None,
        link_options=None,
        vlan_options=None,
        routers=None,
        router_options=None,
        include=None,
        include_optional=None,
        ignored_switches=None,
    ):
        """
        Creates a Faucet YAML configuration file using the current topology

        Args:
            n_vlans (int): Number of VLANs to generate
            acl_options (dict): Acls in use in the Faucet configuration file
            dp_options (dict): Additional options for each DP, keyed by DP index
            host_options (dict): Additional options for each host, keyed by host index
            link_options (dict): Additional options for each link, keyed by indices tuple (u, v)
            vlan_options (dict): Additional options for each VLAN, keyed by vlan index
            routers (dict): Router index to list of VLANs in the router
            router_options (dict): Additional options for each router, keyed by router index
            include (list): Files to include using the the Faucet config 'include' key
            include_optional (list): File to include using the Faucet config 'include_optional' key
            ignored_switches (list): List of switches to not include for the configuration file
        """
        config = {"version": 2}
        if include:
            config["include"] = list(include)
        if include_optional:
            config["include-optional"] = list(include_optional)
        if acl_options:
            config["acls"] = self.get_acls_config(acl_options)
        config["vlans"] = self.get_vlans_config(n_vlans, vlan_options)
        if routers:
            config["routers"] = self.get_routers_config(routers, router_options)
        if ignored_switches is None:
            ignored_switches = []
        config["dps"] = self.get_dps_config(
            dp_options, host_options, link_options, ignored_switches
        )
        return yaml_dump(config)


class FaucetFakeOFTopoGenerator(FaucetTopoGenerator):
    """Generates Faucet topologies for Unittests"""

    # NOTE: For now, we dont actually create the objects for the unittests
    #   so we can leave them as they are in the FaucetTopoGenerator function
    @staticmethod
    def dp_dpid(i):
        """DP DPID"""
        return "%u" % (i + 1)



#!/usr/bin/env python3

"""Implement configuration file parsing."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import re

from faucet import config_parser_util
from faucet.acl import ACL
from faucet.conf import test_config_condition, InvalidConfigError
from faucet.dp import DP
from faucet.meter import Meter
from faucet.port import Port
from faucet.router import Router
from faucet.vlan import VLAN
from faucet.watcher_conf import WatcherConf


V2_TOP_CONFS = ("acls", "dps", "meters", "routers", "vlans")


def dp_parser(config_file, logname, meta_dp_state=None):
    """Parse a config file into DP configuration objects with hashes of config include/files."""
    conf, _ = config_parser_util.read_config(config_file, logname)
    config_hashes = None
    dps = None

    test_config_condition(conf is None, "Config file is empty")
    test_config_condition(
        not isinstance(conf, dict), "Config file does not have valid syntax"
    )
    version = conf.pop("version", 2)
    test_config_condition(version != 2, "Only config version 2 is supported")
    config_hashes, config_contents, dps, top_conf = _config_parser_v2(
        config_file, logname, meta_dp_state
    )
    test_config_condition(dps is None, "no DPs are not defined")

    return config_hashes, config_contents, dps, top_conf


def _get_vlan_by_key(dp_id, vlan_key, vlans):
    try:
        if vlan_key in vlans:
            return vlans[vlan_key]
    except TypeError as err:
        raise InvalidConfigError(err) from err
    for vlan in vlans.values():
        if vlan_key == vlan.vid:
            return vlan
    test_config_condition(
        not isinstance(vlan_key, int),
        (
            "Implicitly created VLAN %s must be an int (not %s)"
            % (vlan_key, type(vlan_key))
        ),
    )
    # Create VLAN with VID, if not defined.
    return vlans.setdefault(vlan_key, VLAN(vlan_key, dp_id))


def _dp_parse_port(dp_id, port_key, port_conf, vlans):
    def _dp_parse_native_port_vlan():
        if port.native_vlan is not None:
            vlan = _get_vlan_by_key(dp_id, port.native_vlan, vlans)
            port.native_vlan = vlan

    def _dp_parse_tagged_port_vlans():
        if port.tagged_vlans:
            port_tagged_vlans = [
                _get_vlan_by_key(dp_id, vlan_key, vlans)
                for vlan_key in port.tagged_vlans
            ]
            port.tagged_vlans = port_tagged_vlans

    port = Port(port_key, dp_id, port_conf)
    test_config_condition(
        str(port_key) not in (str(port.number), port.name),
        ("Port key %s match port name or port number" % port_key),
    )
    _dp_parse_native_port_vlan()
    _dp_parse_tagged_port_vlans()
    return port


def _dp_add_ports(dp, dp_conf, dp_id, vlans):
    ports_conf = dp_conf.get("interfaces", {})
    port_ranges_conf = dp_conf.get("interface_ranges", {})
    # as users can config port VLAN by using VLAN name, we store vid in
    # Port instance instead of VLAN name for data consistency
    test_config_condition(
        not isinstance(ports_conf, dict), ("Invalid syntax in interface config")
    )
    test_config_condition(
        not isinstance(port_ranges_conf, dict),
        ("Invalid syntax in interface ranges config"),
    )

    def _map_port_num_to_port(ports_conf):
        port_num_to_port_conf = {}
        for port_key, port_conf in ports_conf.items():
            test_config_condition(
                not isinstance(port_conf, dict), "Invalid syntax in port config"
            )
            port_num = port_conf.get("number", port_key)
            try:
                port_num_to_port_conf[port_num] = (port_key, port_conf)
            except TypeError as type_error:
                raise InvalidConfigError(
                    "Invalid syntax in port config"
                ) from type_error
        return port_num_to_port_conf

    def _parse_port_ranges(port_ranges_conf, port_num_to_port_conf):
        all_port_nums = set()
        for port_range, port_conf in port_ranges_conf.items():
            # port range format: 1-6 OR 1-6,8-9 OR 1-3,5,7-9
            test_config_condition(
                not isinstance(port_conf, dict), "Invalid syntax in port config"
            )
            port_nums = set()
            if "number" in port_conf:
                del port_conf["number"]
            for range_ in re.findall(r"(\d+-\d+)", str(port_range)):
                start_num, end_num = [int(num) for num in range_.split("-")]
                test_config_condition(
                    start_num >= end_num,
                    ("Incorrect port range (%d - %d)" % (start_num, end_num)),
                )
                port_nums.update(range(start_num, end_num + 1))
                port_range = re.sub(range_, "", port_range)
            other_nums = [int(p) for p in re.findall(r"\d+", str(port_range))]
            port_nums.update(other_nums)
            test_config_condition(
                not port_nums, "interface-ranges contain invalid config"
            )
            test_config_condition(
                port_nums.intersection(all_port_nums),
                "interfaces-ranges cannot overlap",
            )
            all_port_nums.update(port_nums)
            for port_num in port_nums:
                if port_num in port_num_to_port_conf:
                    # port range config has lower priority than individual port config
                    for attr, value in port_conf.items():
                        port_num_to_port_conf[port_num][1].setdefault(attr, value)
                else:
                    port_num_to_port_conf[port_num] = (port_num, port_conf)

    port_num_to_port_conf = _map_port_num_to_port(ports_conf)
    _parse_port_ranges(port_ranges_conf, port_num_to_port_conf)

    for port_num, port_conf in port_num_to_port_conf.values():
        port = _dp_parse_port(dp_id, port_num, port_conf, vlans)
        dp.add_port(port)


def _parse_acls(dp, acls_conf):
    for acl_key, acl_conf in acls_conf.items():
        acl = ACL(acl_key, dp.dp_id, acl_conf)
        dp.add_acl(acl_key, acl)


def _parse_routers(dp, routers_conf):
    for router_key, router_conf in routers_conf.items():
        router = Router(router_key, dp.dp_id, router_conf)
        dp.add_router(router_key, router)


def _parse_meters(dp, meters_conf):
    for meter_key, meter_conf in meters_conf.items():
        meter = Meter(meter_key, dp.dp_id, meter_conf)
        dp.meters[meter_key] = meter


def _parse_dp(dp_key, dp_conf, acls_conf, meters_conf, routers_conf, vlans_conf):
    test_config_condition(not isinstance(dp_conf, dict), "DP config must be dict")
    dp = DP(dp_key, dp_conf.get("dp_id", None), dp_conf)
    test_config_condition(
        dp.name != dp_key, ("DP key %s and DP name must match" % dp_key)
    )
    vlans = {}
    vids = set()
    for vlan_key, vlan_conf in vlans_conf.items():
        vlan = VLAN(vlan_key, dp.dp_id, vlan_conf)
        test_config_condition(
            str(vlan_key) not in (str(vlan.vid), vlan.name),
            ("VLAN %s key must match VLAN name or VLAN VID" % vlan_key),
        )
        test_config_condition(
            not isinstance(vlan_key, (str, int)),
            ("VLAN %s key must not be type %s" % (vlan_key, type(vlan_key))),
        )
        test_config_condition(
            vlan.vid in vids, ("VLAN VID %u multiply configured" % vlan.vid)
        )
        vlans[vlan_key] = vlan
        vids.add(vlan.vid)
    _parse_acls(dp, acls_conf)
    _parse_routers(dp, routers_conf)
    _parse_meters(dp, meters_conf)
    _dp_add_ports(dp, dp_conf, dp.dp_id, vlans)
    return (dp, vlans)


def _dp_parser_v2(
    dps_conf, acls_conf, meters_conf, routers_conf, vlans_conf, meta_dp_state
):
    # pylint: disable=invalid-name
    dp_vlans = []
    for dp_key, dp_conf in dps_conf.items():
        try:
            dp, vlans = _parse_dp(
                dp_key, dp_conf, acls_conf, meters_conf, routers_conf, vlans_conf
            )
            dp_vlans.append((dp, vlans))
        except InvalidConfigError as err:
            raise InvalidConfigError("DP %s: %s" % (dp_key, err)) from err

    # Some VLANs are created implicitly just by referencing them in tagged/native,
    # so we must make them available to all DPs.
    implicit_vids = set()
    for dp, vlans in dp_vlans:
        implicit_vids.update(set(vlans.keys()) - set(vlans_conf.keys()))
    dps = []
    for dp, vlans in dp_vlans:
        for vlan_key in implicit_vids:
            if vlan_key not in vlans:
                vlans[vlan_key] = VLAN(vlan_key, dp.dp_id)
        dp.reset_refs(vlans=vlans)
        dps.append(dp)

    for dp in dps:
        dp.finalize_config(dps)
    for dp in dps:
        dp.resolve_stack_topology(dps, meta_dp_state)
    for dp in dps:
        dp.finalize()

    dpid_refs = set()
    for dp in dps:
        test_config_condition(
            dp.dp_id in dpid_refs, ("DPID %u is duplicated" % dp.dp_id)
        )
        dpid_refs.add(dp.dp_id)

    routers_referenced = set()
    for dp in dps:
        routers_referenced.update(dp.routers.keys())
    for router in routers_conf:
        test_config_condition(
            router not in routers_referenced,
            ("router %s configured but not used by any DP" % router),
        )

    return dps


def dp_preparsed_parser(top_confs, meta_dp_state):
    """Parse a preparsed (after include files have been applied) FAUCET config."""
    local_top_confs = copy.deepcopy(top_confs)
    return _dp_parser_v2(
        local_top_confs.get("dps", {}),
        local_top_confs.get("acls", {}),
        local_top_confs.get("meters", {}),
        local_top_confs.get("routers", {}),
        local_top_confs.get("vlans", {}),
        meta_dp_state,
    )


def _config_parser_v2(config_file, logname, meta_dp_state):
    config_path = config_parser_util.dp_config_path(config_file)
    top_confs = {top_conf: {} for top_conf in V2_TOP_CONFS}
    config_hashes = {}
    config_contents = {}
    dps = None

    if not config_parser_util.dp_include(
        config_hashes, config_contents, config_path, logname, top_confs
    ):
        raise InvalidConfigError(
            "Error found while loading config file: %s" % config_path
        )

    if not top_confs["dps"]:
        raise InvalidConfigError("DPs not configured in file: %s" % config_path)

    dps = dp_preparsed_parser(top_confs, meta_dp_state)
    return (config_hashes, config_contents, dps, top_confs)


def watcher_parser(config_file, logname, prom_client):
    """Return Watcher instances from config."""
    conf, _ = config_parser_util.read_config(config_file, logname)
    conf_hash = config_parser_util.config_file_hash(config_file)
    faucet_config_files, faucet_conf_hashes, result = _watcher_parser_v2(
        conf, logname, prom_client
    )
    return conf_hash, faucet_config_files, faucet_conf_hashes, result


def _parse_dps_for_watchers(conf, logname, meta_dp_state=None):
    all_dps_list = []
    faucet_conf_hashes = {}

    if not isinstance(conf, dict):
        raise InvalidConfigError("Gauge config not valid")

    faucet_config_files = conf.get("faucet_configs", [])
    for faucet_config_file in faucet_config_files:
        conf_hashes, _, dp_list, _ = dp_parser(faucet_config_file, logname)
        if dp_list:
            faucet_conf_hashes[faucet_config_file] = conf_hashes
            all_dps_list.extend(dp_list)

    faucet_config = conf.get("faucet", None)
    if faucet_config:
        all_dps_list.extend(dp_preparsed_parser(faucet_config, meta_dp_state))

    dps = {dp.name: dp for dp in all_dps_list}
    if not dps:
        raise InvalidConfigError("Gauge configured without any FAUCET configuration")
    return faucet_config_files, faucet_conf_hashes, dps


def _watcher_parser_v2(conf, logname, prom_client):
    logger = config_parser_util.get_logger(logname)

    if conf is None:
        conf = {}
    faucet_config_files, faucet_conf_hashes, dps = _parse_dps_for_watchers(
        conf, logname
    )
    dbs = conf.pop("dbs")

    result = []
    for watcher_name, watcher_conf in conf["watchers"].items():
        if watcher_conf.get("all_dps", False):
            watcher_dps = dps.keys()
        else:
            watcher_dps = watcher_conf["dps"]
        # Watcher config has a list of DPs, but actually a WatcherConf is
        # created for each DP.
        # TODO: refactor watcher_conf as a container.
        for dp_name in watcher_dps:
            if dp_name not in dps:
                logger.error("DP %s in Gauge but not configured in FAUCET", dp_name)
                continue
            dp = dps[dp_name]
            if "dbs" in watcher_conf:
                watcher_dbs = watcher_conf["dbs"]
            elif "db" in watcher_conf:
                watcher_dbs = [watcher_conf["db"]]
            else:
                raise InvalidConfigError("Watcher configured without DB")
            for db in watcher_dbs:
                watcher = WatcherConf(watcher_name, dp.dp_id, watcher_conf, prom_client)
                watcher.add_db(dbs[db])
                watcher.add_dp(dp)
                result.append(watcher)

    return faucet_config_files, faucet_conf_hashes, result


def get_config_for_api(valves):
    """Return config as dict for all DPs."""
    config = {i: {} for i in V2_TOP_CONFS}
    for valve in valves.values():
        valve_conf = valve.get_config_dict()
        for i in V2_TOP_CONFS:
            if i in valve_conf:
                config[i].update(valve_conf[i])  # pytype: disable=attribute-error
    return config



"""Unit tests for Port"""

import unittest

from faucet.port import Port
from faucet.port import (
    LACP_ACTOR_NOTCONFIGURED,
    LACP_PORT_NOTCONFIGURED,
    LACP_ACTOR_INIT,
    LACP_ACTOR_UP,
    LACP_ACTOR_NOSYNC,
    LACP_ACTOR_NONE,
    LACP_PORT_UNSELECTED,
    LACP_PORT_SELECTED,
    LACP_PORT_STANDBY,
)


class MockVLAN:  # pylint: disable=too-few-public-methods
    """Mock class for VLAN so we can inject into Port"""

    def __init__(self, name):
        self.name = name


class FaucetPortMethodTest(unittest.TestCase):  # pytype: disable=module-attr
    """Test a range of methods on Port"""

    def test_vlans(self):
        """Test that the vlans() method behaves correctly"""

        vlan100 = MockVLAN("v100")
        vlan200 = MockVLAN("v200")
        vlan300 = MockVLAN("v300")

        tagged_vlans = [vlan200, vlan300]
        native_vlan = vlan100

        port = Port(1, 1, {})
        port.native_vlan = native_vlan
        self.assertIn(native_vlan, port.vlans())
        port.tagged_vlans = tagged_vlans
        self.assertEqual(set(port.vlans()), set([native_vlan] + tagged_vlans))
        port.native_vlan = None
        self.assertEqual(set(port.vlans()), set(tagged_vlans))


class FaucetLACPPortFunctions(unittest.TestCase):  # pytype: disable=module-attr
    """Test port LACP state functions work as expected"""

    def test_lacp_actor_update(self):
        """Test updating port LACP information causes correct actor state changes"""
        port = Port(1, 1, {})
        port.enabled = True
        port.dyn_phys_up = True
        # Before configuring the LACP state machine, default is notconfigured
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_NOTCONFIGURED)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_NOTCONFIGURED)
        # Move to initial configuration state when no packets have been received yet
        port.lacp_actor_update(False, None, None)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_INIT)
        # Receiving first packets but no SYNC bit set
        port.lacp_actor_update(False, 1, 1)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_NOSYNC)
        # Receiving sync packets
        port.lacp_actor_update(True, 1, 1)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_UP)
        # Port phys down, move to ACTOR_NONE state
        port.dyn_phys_up = False
        port.lacp_actor_update(True, 1, 1)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_NONE)
        port.lacp_actor_update(False, None, None)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_NONE)
        # Go back to init once port restarted
        port.dyn_phys_up = True
        port.lacp_actor_update(False, None, None)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_INIT)
        # Cold starting will force the port to revert to notconfigured state
        port.lacp_actor_update(False, None, None, True)
        self.assertEqual(port.dyn_lacp_actor_state, LACP_ACTOR_NOTCONFIGURED)

    def test_lacp_flags(self):
        """Test port LACP flags returns correct flags for current port states"""
        port = Port(1, 1, {})
        # LACP config option to force flags on
        port.lacp_collect_and_distribute = True
        self.assertEqual(port.get_lacp_flags(), (1, 1, 1))
        port.lacp_collect_and_distribute = False
        # Port is selected, so flags should be on
        port.dyn_lacp_port_selected = LACP_PORT_SELECTED
        self.assertEqual(port.get_lacp_flags(), (1, 1, 1))
        # Port in standby, only allow sync
        port.dyn_lacp_port_selected = LACP_PORT_STANDBY
        self.assertEqual(port.get_lacp_flags(), (1, 0, 0))
        # Port not in standby, or selected
        port.dyn_lacp_port_selected = LACP_PORT_UNSELECTED
        self.assertEqual(port.get_lacp_flags(), (0, 0, 0))
        port.dyn_lacp_port_selected = LACP_PORT_NOTCONFIGURED
        self.assertEqual(port.get_lacp_flags(), (0, 0, 0))

    def test_lacp_port_update(self):
        """Test LACP port options"""
        port = Port(1, 1, {})
        port.enabled = True
        port.dyn_phys_up = True
        # Port defaults to NOTCONFIGURED
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_NOTCONFIGURED)
        # Now call update to configure port and get initial state
        #   depending on the chosen DP
        port.lacp_port_update(False)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_UNSELECTED)
        # Now select our the port's current DP
        port.lacp_port_update(True)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_SELECTED)
        # Test option to force standby mode
        # Option forces the statemachine to revert to STANDBY mode when not selected
        port.lacp_standby = True
        port.lacp_port_update(True)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_SELECTED)
        port.lacp_port_update(False)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_STANDBY)
        # Test forcing selected port
        port.lacp_standby = False
        port.lacp_selected = True
        port.lacp_port_update(False)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_SELECTED)
        # Test forcing unselected port
        port.lacp_selected = False
        port.lacp_unselected = True
        port.lacp_port_update(True)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_UNSELECTED)
        # Test reverting to unconfigured on port cold start
        port.lacp_unselected = False
        port.lacp_port_update(False, True)
        self.assertEqual(port.dyn_lacp_port_selected, LACP_PORT_NOTCONFIGURED)


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""Gauge watcher implementations."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import gzip
import time

from os_ken.ofproto import ofproto_v1_3 as ofp

from faucet.conf import InvalidConfigError
from faucet.valve_util import dpid_log
from faucet.gauge_influx import (
    GaugePortStateInfluxDBLogger,
    GaugePortStatsInfluxDBLogger,
    GaugeFlowTableInfluxDBLogger,
)
from faucet.gauge_pollers import (
    GaugePortStatePoller,
    GaugePortStatsPoller,
    GaugeFlowTablePoller,
    GaugeMeterStatsPoller,
)
from faucet.gauge_prom import (
    GaugePortStatsPrometheusPoller,
    GaugePortStatePrometheusPoller,
    GaugeFlowTablePrometheusPoller,
    GaugeMeterStatsPrometheusPoller,
)


def watcher_factory(conf):
    """Return a Gauge object based on type.

    Args:
        conf (GaugeConf): object with the configuration for this valve.
    """

    watcher_types = {
        "port_state": {
            "text": GaugePortStateLogger,
            "influx": GaugePortStateInfluxDBLogger,
            "prometheus": GaugePortStatePrometheusPoller,
        },
        "port_stats": {
            "text": GaugePortStatsLogger,
            "influx": GaugePortStatsInfluxDBLogger,
            "prometheus": GaugePortStatsPrometheusPoller,
        },
        "flow_table": {
            "text": GaugeFlowTableLogger,
            "influx": GaugeFlowTableInfluxDBLogger,
            "prometheus": GaugeFlowTablePrometheusPoller,
        },
        "meter_stats": {
            "text": GaugeMeterStatsLogger,
            "prometheus": GaugeMeterStatsPrometheusPoller,
        },
    }

    w_type = conf.type
    db_type = conf.db_type
    try:
        return watcher_types[w_type][db_type]
    except KeyError as key_error:
        raise InvalidConfigError("invalid water config") from key_error


class GaugePortStateLogger(GaugePortStatePoller):
    """Abstraction for port state logger."""

    def _update(self, rcv_time, msg):
        rcv_time_str = self._rcv_time(rcv_time)
        reason = msg.reason
        port_no = msg.desc.port_no
        log_msg = "port %s unknown state %s" % (port_no, reason)
        if reason == ofp.OFPPR_ADD:
            log_msg = "port %s added" % port_no
        elif reason == ofp.OFPPR_DELETE:
            log_msg = "port %s deleted" % port_no
        elif reason == ofp.OFPPR_MODIFY:
            link_down = msg.desc.state & ofp.OFPPS_LINK_DOWN
            if link_down:
                log_msg = "port %s down" % port_no
            else:
                log_msg = "port %s up" % port_no
        log_msg = "%s %s" % (dpid_log(self.dp.dp_id), log_msg)
        self.logger.info(log_msg)
        if self.conf.file:
            with open(self.conf.file, "a", encoding="utf-8") as logfile:
                logfile.write("\t".join((rcv_time_str, log_msg)) + "\n")

    def send_req(self):
        """Send a stats request to a datapath."""
        raise NotImplementedError  # pragma: no cover

    def no_response(self):
        """Called when a polling cycle passes without receiving a response."""
        raise NotImplementedError  # pragma: no cover


class GaugePortStatsLogger(GaugePortStatsPoller):
    """Abstraction for port statistics logger."""

    def _dp_stat_name(self, stat, stat_name):
        port_name = self.dp.port_labels(stat.port_no)["port"]
        return "-".join((self.dp.name, port_name, stat_name))


class GaugeMeterStatsLogger(GaugeMeterStatsPoller):
    """Abstraction for meter statistics logger."""

    def _format_stat_pairs(self, delim, stat):
        band_stats = stat.band_stats[0]
        stat_pairs = (
            (("flow", "count"), stat.flow_count),
            (("byte", "in", "count"), stat.byte_in_count),
            (("packet", "in", "count"), stat.packet_in_count),
            (("byte", "band", "count"), band_stats.byte_band_count),
            (("packet", "band", "count"), band_stats.packet_band_count),
        )
        return self._format_stats(delim, stat_pairs)

    def _dp_stat_name(self, stat, stat_name):
        return "-".join((self.dp.name, str(stat.meter_id), stat_name))


class GaugeFlowTableLogger(GaugeFlowTablePoller):
    """Periodically dumps the current datapath flow table as a yaml object.

    Includes a timestamp and a reference ($DATAPATHNAME-flowtables). The
    flow table is dumped as an OFFlowStatsReply message (in yaml format) that
    matches all flows.

    optionally the output can be compressed by setting compressed: true in the
    config for this watcher
    """

    def _rcv_time(self, rcv_time):
        # Use ISO8601 times for filenames
        return time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(rcv_time))

    def _update(self, rcv_time, msg):
        rcv_time_str = self._rcv_time(rcv_time)
        path = self.conf.path
        # Double Hyphen to avoid confusion with ISO8601 times
        filename = os.path.join(
            path, "{}--flowtable--{}.json".format(self.dp.name, rcv_time_str)
        )
        if os.path.isfile(filename):
            # If this filename already exists, add an increment to the filename
            # (for dealing with parts of a multipart message arriving at the same time)
            inc = 1
            while os.path.isfile(filename):
                filename = os.path.join(
                    path,
                    "{}--flowtable--{}--{}.json".format(
                        self.dp.name, rcv_time_str, inc
                    ),
                )
                inc += 1

        if self.conf.compress:
            with gzip.open(filename, "wt") as outfile:
                outfile.write(json.dumps(msg.to_jsondict()))
        else:
            with open(filename, "w", encoding="utf-8") as outfile:
                json.dump(msg.to_jsondict(), outfile, indent=2)



"""Implementation of Valve coprocessor."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet import valve_of
from faucet.valve_manager_base import ValveManagerBase
from faucet.vlan import OFVLAN


class CoprocessorManager(ValveManagerBase):
    """Implementation of Valve coprocessor."""

    def __init__(
        self,
        ports,
        copro_table,
        vlan_table,
        eth_src_table,  # pylint: disable=too-many-arguments
        output_table,
        low_priority,
        high_priority,
    ):
        self.ports = ports
        self.copro_table = copro_table
        self.vlan_table = vlan_table
        self.eth_src_table = eth_src_table
        self.output_table = output_table
        self.low_priority = low_priority
        self.high_priority = high_priority

    def add_port(self, port):
        """Add flows to allow coprocessor to inject or output packets."""
        ofmsgs = []
        if port.coprocessor:
            ofmsgs.append(
                self.vlan_table.flowmod(
                    self.vlan_table.match(in_port=port.number),
                    priority=self.low_priority,
                    inst=(self.vlan_table.goto(self.copro_table),),
                )
            )
            ofmsgs.append(
                self.eth_src_table.flowmod(
                    match=self.eth_src_table.match(in_port=port.number),
                    priority=self.high_priority,
                    inst=(self.eth_src_table.goto(self.output_table),),
                )
            )
            # TODO: add additional output port strategies (eg. MPLS) and tagged ports
            vlan_vid_base = port.coprocessor.get("vlan_vid_base", 0)
            for port_number in self.ports:
                inst = (
                    valve_of.apply_actions(
                        (valve_of.pop_vlan(), valve_of.output_port(port_number))
                    ),
                )
                vid = vlan_vid_base + port_number
                vlan = OFVLAN(str(vid), vid)
                match = self.copro_table.match(vlan=vlan)
                ofmsgs.append(
                    self.copro_table.flowmod(
                        match=match, priority=self.high_priority, inst=inst
                    )
                )
        return ofmsgs



#!/usr/bin/env python3

"""Test fctl FAUCET CLI utility."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
import shutil
import subprocess
import tempfile
import unittest

from faucet import fctl


class FctlTestCaseBase(unittest.TestCase):  # pytype: disable=module-attr
    """Base class for fctl tests."""

    DEFAULT_VALUES = {
        "dp_id": "0xb827eb608918",
        "mac_addr": "a4:5e:60:c5:5c:ed",
        "metrics": "learned_macs",
        "n": 3,
        "port": "17",
        "vlan": "2004",
        "value": 180725257428205.0,
    }

    SRC_DIR = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "../../../faucet"
    )

    FCTL_BASE_ARGS = [
        "--metrics={metrics}".format(**DEFAULT_VALUES),
        "--labels=dp_id:{dp_id}".format(**DEFAULT_VALUES),
    ]
    FCTL = os.path.join(SRC_DIR, "fctl.py")
    tmpdir = None
    prom_input_file_name = None

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.prom_input_file_name = os.path.join(self.tmpdir, "prom_input.txt")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def fctl_args(self, extra_args=None):
        """generate argument list for fctl"""
        result = copy.copy(self.FCTL_BASE_ARGS)
        result += ["--endpoints=file:%s" % self.prom_input_file_name]
        if extra_args is not None:
            result += extra_args
        return result

    def learned_macs_prom(self, overwrite_labels=None):
        """generate prometheus formated data"""
        labels = copy.copy(self.DEFAULT_VALUES)
        if overwrite_labels is not None:
            labels.update(overwrite_labels)
        result = """
{metrics}{{dp_id="{dp_id}",n="{n}",port="{port}",vlan="{vlan}"}}\t{value}"""
        return result.format(**labels).strip()

    def learned_macs_result(self, overwrite_labels=None):
        """generate expected output data"""
        labels = copy.copy(self.DEFAULT_VALUES)
        if overwrite_labels is not None:
            labels.update(overwrite_labels)
        result = """
{metrics}\t[('dp_id', '{dp_id}'), ('n', '{n}'), ('port', '{port}'), ('vlan', '{vlan}')]\t{mac_addr}
"""
        return result.format(**labels).strip()


class FctlTestCase(FctlTestCaseBase):
    """Drive fctl from shell."""

    def run_fctl(self, prom_input, expected_output, extra_args=None):
        """Ensure fctl succeeds and returns expected output."""
        with open(self.prom_input_file_name, "w", encoding="utf-8") as prom_input_file:
            prom_input_file.write(prom_input)
        fctl_cli = " ".join(["python3", self.FCTL] + self.fctl_args(extra_args))
        retcode, output = subprocess.getstatusoutput(
            fctl_cli
        )  # pytype: disable=module-attr
        self.assertEqual(0, retcode, msg="%s returned %d" % (fctl_cli, retcode))
        output = output.strip()
        self.assertEqual(output, expected_output)

    def test_macs(self):
        """Test can parse learned MACs from Prometheus data."""
        self.run_fctl(self.learned_macs_prom(), self.learned_macs_result())

    def test_display_labels(self):
        """Test can filter by display labels."""
        expected_output = """
learned_macs\t[('dp_id', '{dp_id}')]\t{mac_addr}
""".format(
            **self.DEFAULT_VALUES
        ).strip()

        self.run_fctl(
            self.learned_macs_prom(),
            expected_output,
            extra_args=["--display-labels=dp_id"],
        )


class FctlClassTestCase(FctlTestCaseBase):
    """Test fctl internal methods."""

    def test_http_fail(self):
        """Test HTTP scrape handled."""
        with open(os.devnull, "w", encoding="utf-8") as err_output_file:
            self.assertEqual(
                None,
                fctl.scrape_prometheus(
                    ["http://127.0.0.1:23"], err_output_file=err_output_file
                ),
            )

    def test_bad_url(self):
        """Test unparseable URL."""
        with open(os.devnull, "w", encoding="utf-8") as err_output_file:
            self.assertEqual(
                None,
                fctl.scrape_prometheus(
                    ["not/a$#@/valid_URL"], err_output_file=err_output_file
                ),
            )

    def test_bad_content(self):
        """Test bad content."""
        bad_input_file_name = os.path.join(self.tmpdir, "bad_content.txt")
        with open(bad_input_file_name, "w", encoding="utf-8") as bad_input_file:
            bad_input_file.write("NOT/_prometheus_data")
        with open(os.devnull, "w", encoding="utf-8") as err_output_file:
            self.assertEqual(
                None,
                fctl.scrape_prometheus(
                    ["file://%s" % bad_input_file_name], err_output_file=err_output_file
                ),
            )

    def write_prom_input_file(self, input_data):
        with open(self.prom_input_file_name, "w", encoding="utf-8") as prom_input_file:
            prom_input_file.write(input_data)

    def test_macs(self):
        """Test reporting of learned MACs."""
        self.write_prom_input_file(self.learned_macs_prom())
        (endpoints, report_metrics, label_matches, nonzero_only, _) = fctl.parse_args(
            self.fctl_args()
        )
        metrics = fctl.scrape_prometheus(endpoints)
        report_out = fctl.report_label_match_metrics(
            report_metrics=report_metrics,
            metrics=metrics,
            label_matches=label_matches,
            nonzero_only=nonzero_only,
        )
        self.assertEqual(report_out, self.learned_macs_result())

    def test_get_samples(self):
        """Test querying with get_samples"""
        self.write_prom_input_file(self.learned_macs_prom())
        samples = fctl.get_samples(
            ["file://" + self.prom_input_file_name], "learned_macs", {}
        )
        self.assertEqual(samples[0].value, self.DEFAULT_VALUES["value"])


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""Manage a collection of Valves."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from faucet.conf import InvalidConfigError
from faucet.config_parser_util import config_changed, CONFIG_HASH_FUNC
from faucet.config_parser import dp_parser, dp_preparsed_parser
from faucet.valve import valve_factory, SUPPORTED_HARDWARE
from faucet.valve_util import dpid_log, stat_config_files


class MetaDPState:
    """Contains state/config about all DPs."""

    def __init__(self):
        self.stack_root_name = None
        self.dp_last_live_time = {}
        self.top_conf = None
        self.last_good_config = {}
        self.config_hash_info = {}


class ConfigWatcher:
    """Watch config for file or content changes."""

    config_file = None
    config_hashes = None
    config_file_stats = None

    def files_changed(self):
        """Return True if any config files changed."""
        # TODO: Better to use an inotify method that doesn't conflict with eventlets.
        changed = False
        if self.config_hashes:
            new_config_file_stats = stat_config_files(self.config_hashes)
            if self.config_file_stats:
                # Check content as well in case mtime et al was cached.
                if new_config_file_stats == self.config_file_stats:
                    changed = self.content_changed(self.config_file)
                else:
                    changed = True
            self.config_file_stats = new_config_file_stats
        return changed

    def content_changed(self, new_config_file):
        """Return True if config file content actually changed."""
        return config_changed(self.config_file, new_config_file, self.config_hashes)

    def update(self, new_config_file, new_config_hashes=None):
        """Update state with new config file/hashes."""
        self.config_file = new_config_file
        if new_config_hashes is None:
            new_config_hashes = {new_config_file: None}
        if new_config_hashes:
            self.config_hashes = new_config_hashes


class ValvesManager:
    """Manage a collection of Valves."""

    def __init__(
        self,
        logname,
        logger,
        metrics,
        notifier,
        bgp,
        dot1x,
        config_auto_revert,
        send_flows_to_dp_by_id,
    ):
        """Initialize ValvesManager.

        Args:
            logname (str): log name to use in logging.
            logger (logging.logging): logger instance to use for logging.
            metrics (FaucetMetrics): metrics instance.
            notifier (FaucetEvent): event notifier instance.
            bgp (FaucetBgp): BGP instance.
            config_auto_revert (bool): True if FAUCET should attempt to revert bad configs.
            send_flows_to_dp_by_id: callable, two args - DP ID and list of flows to send to DP.
        """
        self.logname = logname
        self.logger = logger
        self.metrics = metrics
        self.notifier = notifier
        self.bgp = bgp
        self.dot1x = dot1x
        self.config_auto_revert = config_auto_revert
        self.send_flows_to_dp_by_id = send_flows_to_dp_by_id
        self.valves = {}
        self.config_applied = {}
        self.config_watcher = ConfigWatcher()
        self.meta_dp_state = MetaDPState()

    def update_dp_live_time(self, now):
        """
        Update DP running time

        Args:
            now (float): Current time
        """
        for valve in self.valves.values():
            if valve.dp.dyn_running:
                self.meta_dp_state.dp_last_live_time[valve.dp.name] = now

    def reload_stack_root_config(self, now):
        """
        Force reload & apply configuration for stack root changes
        Args:
            now (float): Current time
        """
        dps = dp_preparsed_parser(self.meta_dp_state.top_conf, self.meta_dp_state)
        self._apply_configs(dps, now, None)

    def valves_by_name(self):
        """Return a name/valve dict of all the stacking valves"""
        return {
            valve.dp.name: valve
            for valve in self.valves.values()
            if valve.stack_manager
        }

    def maintain_stack_root(self, now, update_time):
        """
        Maintain current stack root

        Args:
            now (float): Current time
            update_time (int): Stack root update time interval
        """
        self.update_dp_live_time(now)
        last_live_times = self.meta_dp_state.dp_last_live_time

        valves_by_name = self.valves_by_name()
        if not valves_by_name:
            return False

        prev_root_name = self.meta_dp_state.stack_root_name
        prev_root_valve = valves_by_name.get(prev_root_name, None)

        prev_other_valves = [
            valve for valve in self.valves.values() if valve != prev_root_valve
        ]

        new_root_name = list(valves_by_name.values())[
            0
        ].stack_manager.nominate_stack_root(
            prev_root_valve, prev_other_valves, now, last_live_times, update_time
        )
        return self.set_stack_root(now, new_root_name)

    def set_stack_root(self, now, new_root_name):
        """
        Set stack root

        Args:
            now (float): Current time
            new_root_name (string): Name of new stack root
        """
        valves_by_name = self.valves_by_name()
        prev_root_name = self.meta_dp_state.stack_root_name
        prev_root_valve = valves_by_name.get(prev_root_name, None)
        new_root_valve = valves_by_name.get(new_root_name, None)

        stack_change = False
        if new_root_valve:
            if prev_root_name != new_root_name:
                # Current stack root is not the new stack root
                self.logger.info(
                    "Stack root %s (previous %s)" % (new_root_name, prev_root_name)
                )
                if prev_root_valve:
                    labels = prev_root_valve.dp.base_prom_labels()
                    self.metrics.is_dp_stack_root.labels(**labels).set(0)
                self.meta_dp_state.stack_root_name = new_root_name
                self.metrics.faucet_stack_root_dpid.set(new_root_valve.dp.dp_id)
                self.metrics.stack_root_change_count.inc(1)
                self.reload_stack_root_config(now)
                if prev_root_name:
                    stack_change = True
            else:
                # Current stack root does not change, however ensure that the current stack root
                #   is known for all DPs
                new_other_valves = [
                    valve for valve in self.valves.values() if valve != new_root_valve
                ]
                inconsistent_dps = not new_root_valve.stack_manager.consistent_roots(
                    prev_root_name, new_root_valve, new_other_valves
                )
                if inconsistent_dps:
                    self.logger.info("Reloading stack DPs (stack root inconsistent)")
                    self.reload_stack_root_config(now)
                    stack_change = True
            labels = new_root_valve.dp.base_prom_labels()
            self.metrics.is_dp_stack_root.labels(**labels).set(1)
            for valve in valves_by_name.values():
                labels = valve.dp.base_prom_labels()
                path_port = valve.stack_manager.chosen_towards_port
                path_port_number = path_port.number if path_port else 0.0
                self.metrics.dp_root_hop_port.labels(**labels).set(path_port_number)
        if stack_change:
            for valve in valves_by_name.values():
                valve.stale_root = True
        return stack_change

    def event_socket_heartbeat(self):
        """raises event for event sock heartbeat"""
        self._notify({"EVENT_SOCK_HEARTBEAT": None})

    def revert_config(self):
        """Attempt to revert config to last known good version."""
        for (
            config_file_name,
            config_content,
        ) in self.meta_dp_state.last_good_config.items():
            self.logger.info(
                "attempting to revert to last good config: %s" % config_file_name
            )
            try:
                with open(config_file_name, "w", encoding="utf-8") as config_file:
                    config_file.write(str(config_content))
            except (FileNotFoundError, OSError, PermissionError) as err:
                self.logger.error("could not revert %s: %s" % (config_file_name, err))
                return
        self.logger.info("successfully reverted to last good config")

    def parse_configs(self, new_config_file):
        """Return parsed configs for Valves, or None."""
        self.metrics.faucet_config_hash_func.labels(algorithm=CONFIG_HASH_FUNC)
        try:
            new_conf_hashes, new_config_content, new_dps, top_conf = dp_parser(
                new_config_file, self.logname, self.meta_dp_state
            )
            new_present_conf_hashes = [
                (conf_file, conf_hash)
                for conf_file, conf_hash in sorted(new_conf_hashes.items())
                if conf_hash is not None
            ]
            conf_files = [conf_file for conf_file, _ in new_present_conf_hashes]
            conf_hashes = [conf_hash for _, conf_hash in new_present_conf_hashes]
            self.config_watcher.update(new_config_file, new_conf_hashes)
            self.meta_dp_state.top_conf = top_conf
            self.meta_dp_state.last_good_config = new_config_content
            self.meta_dp_state.config_hash_info = dict(
                config_files=",".join(conf_files),
                hashes=",".join(conf_hashes),
                error="",
            )
            self.metrics.faucet_config_hash.info(self.meta_dp_state.config_hash_info)
            self.metrics.faucet_config_load_error.set(0)
        except InvalidConfigError as err:
            self.logger.error("New config bad (%s) - rejecting", err)
            # If the config was reverted, let the watcher notice.
            if self.config_auto_revert:
                self.revert_config()
            self.config_watcher.update(new_config_file)
            self.meta_dp_state.config_hash_info = dict(
                config_files=new_config_file, hashes="", error=str(err)
            )
            self.metrics.faucet_config_hash.info(self.meta_dp_state.config_hash_info)
            self.metrics.faucet_config_load_error.set(1)
            new_dps = None
        return new_dps

    def new_valve(self, new_dp):
        valve_cl = valve_factory(new_dp)
        if valve_cl is not None:
            return valve_cl(
                new_dp, self.logname, self.metrics, self.notifier, self.dot1x
            )
        self.logger.error(
            "%s hardware %s must be one of %s",
            new_dp.name,
            new_dp.hardware,
            sorted(list(SUPPORTED_HARDWARE.keys())),
        )
        return None

    def _apply_configs(self, new_dps, now, delete_dp):
        self.update_config_applied(reset=True)
        if new_dps is None:
            return False
        deleted_dpids = set(self.valves) - {dp.dp_id for dp in new_dps}
        sent = {}
        for new_dp in new_dps:
            dp_id = new_dp.dp_id
            if dp_id in self.valves:
                self.logger.info("Reconfiguring existing datapath %s", dpid_log(dp_id))
                valve = self.valves[dp_id]
                ofmsgs = valve.reload_config(now, new_dp, list(self.valves.values()))
                self.send_flows_to_dp_by_id(valve, ofmsgs)
                sent[dp_id] = valve.dp.dyn_running
            else:
                self.logger.info("Add new datapath %s", dpid_log(new_dp.dp_id))
                valve = self.new_valve(new_dp)
                if valve is None:
                    continue
                self._notify({"CONFIG_CHANGE": {"restart_type": "new"}}, dp=new_dp)
            valve.update_config_metrics()
            self.valves[dp_id] = valve
        if delete_dp is not None:
            for deleted_dp in deleted_dpids:
                delete_dp(deleted_dp)
                del self.valves[deleted_dp]
        self.bgp.reset(self.valves)
        self.dot1x.reset(self.valves)
        self.update_config_applied(sent)
        return True

    def load_configs(self, now, new_config_file, delete_dp=None):
        """Load/apply new config to all Valves."""
        return self._apply_configs(self.parse_configs(new_config_file), now, delete_dp)

    def _send_ofmsgs_by_valve(self, ofmsgs_by_valve):
        if ofmsgs_by_valve:
            for valve, ofmsgs in ofmsgs_by_valve.items():
                self.send_flows_to_dp_by_id(valve, ofmsgs)

    def _notify(self, event_dict, dp=None):
        """Send an event notification."""
        if dp:
            self.notifier.notify(dp.dp_id, dp.name, event_dict)
        else:
            self.notifier.notify(0, str(0), event_dict)

    def request_reload_configs(self, now, new_config_file, delete_dp=None):
        """Process a request to load config changes."""
        if self.config_watcher.content_changed(new_config_file):
            self.logger.info(
                "configuration %s changed, analyzing differences", new_config_file
            )
            result = self.load_configs(now, new_config_file, delete_dp=delete_dp)
            self._notify(
                {
                    "CONFIG_CHANGE": {
                        "success": result,
                        "config_hash_info": self.meta_dp_state.config_hash_info,
                    }
                }
            )
        else:
            self.logger.info("configuration is unchanged, not reloading")
            self.metrics.faucet_config_load_error.set(0)
        self.metrics.faucet_config_reload_requests.inc()  # pylint: disable=no-member

    def update_metrics(self, now):
        """Update metrics in all Valves."""
        for valve in self.valves.values():
            valve.update_metrics(now, rate_limited=False)
        self.bgp.update_metrics(now)

    def valve_flow_services(self, now, valve_service):
        """Call a method on all Valves and send any resulting flows."""
        ofmsgs_by_valve = defaultdict(list)
        for valve in self.valves.values():
            other_valves = self._other_running_valves(valve)
            valve_service_labels = dict(
                valve.dp.base_prom_labels(), valve_service=valve_service
            )
            valve_service_func = getattr(valve, valve_service)
            with self.metrics.faucet_valve_service_secs.labels(  # pylint: disable=no-member
                **valve_service_labels
            ).time():
                for service_valve, ofmsgs in valve_service_func(
                    now, other_valves
                ).items():
                    # Since we are calling all Valves, keep only the ofmsgs
                    # provided by the last Valve called (eventual consistency).
                    if service_valve in ofmsgs_by_valve:
                        ofmsgs_by_valve[service_valve] = []
                    ofmsgs_by_valve[service_valve].extend(ofmsgs)
        self._send_ofmsgs_by_valve(ofmsgs_by_valve)

    def _other_running_valves(self, valve):
        return [
            other_valve
            for other_valve in self.valves.values()
            if valve != other_valve and other_valve.dp.dyn_running
        ]

    def port_desc_stats_reply_handler(self, valve, msg, now):
        """Handle a port desc stats reply message."""
        ofmsgs_by_valve = valve.port_desc_stats_reply_handler(
            msg.body, self._other_running_valves(valve), now
        )
        self._send_ofmsgs_by_valve(ofmsgs_by_valve)

    def port_status_handler(self, valve, msg, now):
        """Handle a port status change message."""
        ofmsgs_by_valve = valve.port_status_handler(
            msg.desc.port_no,
            msg.reason,
            msg.desc.state,
            self._other_running_valves(valve),
            now,
        )
        self._send_ofmsgs_by_valve(ofmsgs_by_valve)

    def valve_packet_in(self, now, valve, msg):
        """Time a call to Valve packet in handler."""
        self.metrics.of_packet_ins.labels(  # pylint: disable=no-member
            **valve.dp.base_prom_labels()
        ).inc()
        if valve.rate_limit_packet_ins(now):
            return
        pkt_meta = valve.parse_pkt_meta(msg)
        if pkt_meta is None:
            self.metrics.of_unexpected_packet_ins.labels(  # pylint: disable=no-member
                **valve.dp.base_prom_labels()
            ).inc()
            return
        with self.metrics.faucet_packet_in_secs.labels(  # pylint: disable=no-member
            **valve.dp.base_prom_labels()
        ).time():
            ofmsgs_by_valve = valve.rcv_packet(
                now, self._other_running_valves(valve), pkt_meta
            )
        if ofmsgs_by_valve:
            self._send_ofmsgs_by_valve(ofmsgs_by_valve)
            valve.update_metrics(now, pkt_meta.port, rate_limited=True)

    def update_config_applied(self, sent=None, reset=False):
        """Update faucet_config_applied from {dpid: sent} dict,
        defining applied == sent == enqueued via Ryu"""
        if reset:
            self.config_applied = defaultdict(bool)
        if sent:
            self.config_applied.update(sent)
        count = float(len(self.valves))
        configured = sum(
            (1 if self.config_applied[dp_id] else 0) for dp_id in self.valves
        )
        fraction = configured / count if count > 0 else 0
        self.metrics.faucet_config_applied.set(fraction)

    def datapath_connect(self, now, valve, discovered_up_ports):
        """Handle connection from DP."""
        self.meta_dp_state.dp_last_live_time[valve.dp.name] = now
        self.update_config_applied({valve.dp.dp_id: True})
        return valve.datapath_connect(now, discovered_up_ports)



"""Base class for managing integration network topologies"""

import os
import random
import re
import time
import ipaddress
import pprint
from functools import partial

from clib.mininet_test_util import timeout_cmd
from clib.mininet_test_base import FaucetTestBase, IPV4_ETH
from clib.config_generator import FaucetTopoGenerator
from clib.valve_test_lib import yaml_load


class FaucetTopoTestBase(FaucetTestBase):
    """
    Extension to the base test for the integration test suite to help set up arbitrary topologies
    """

    NUM_FAUCET_CONTROLLERS = 2
    NUM_GAUGE_CONTROLLERS = 1

    DESCENDING_DPIDS = False

    NETPREFIX = 24
    IPV = 4
    GROUP_TABLE = False

    NUM_DPS = 2
    NUM_HOSTS = 4
    LINKS_PER_HOST = 1

    LACP_TIMEOUT = 10

    dpids = None
    port_maps = None

    n_vlans = 0
    configuration_options = None
    mininet_host_options = None
    host_vlans = None
    link_vlans = None

    host_information = None
    faucet_vips = None

    host_port_maps = {}
    link_port_maps = {}

    def _init_faucet_config(self):
        """Initialize & normalize faucet configuration file"""
        config_vars = {}
        for config_var in (self.config_ports, self.port_map):
            config_vars.update(config_var)
        faucet_config = self.CONFIG % config_vars
        self._write_yaml_conf(self.faucet_config_path, yaml_load(faucet_config))

    def _annotate_interfaces_conf(self, yaml_conf):
        """We don't need to annotate the interfaces"""
        return yaml_conf

    def _dp_ports(self):
        """Return ports on the first DP"""
        return list(self.topo.ports[self.topo.switches_by_id[0]].keys())

    def get_gauge_watcher_config(self):
        """Return gauge watcher config"""
        return """
    port_stats:
        dps: ['%s']
        type: 'port_stats'
        interval: 5
        db: 'stats_file'
    port_state:
        dps: ['%s']
        type: 'port_state'
        interval: 5
        db: 'state_file'
    flow_table:
        dps: ['%s']
        type: 'flow_table'
        interval: 5
        db: 'flow_dir'
""" % (
            self.topo.switches_by_id[0],
            self.topo.switches_by_id[0],
            self.topo.switches_by_id[0],
        )

    def first_switch(self):
        """Return the first switch"""
        return self.net.get(self.topo.switches_by_id[0])

    def port_labels(self, port_no):
        """Return regex for port label"""
        port_name = "b%u" % port_no
        return {"port": port_name, "port_description": r".+"}

    @staticmethod
    def acls():
        """Defined configuration ACLs"""
        return {}

    def faucet_vip(self, i):
        """Faucet VLAN VIP"""
        return "10.%u.0.254/%u" % (i + 1, self.NETPREFIX)

    @staticmethod
    def faucet_mac(i):
        """Faucet VLAN MAC"""
        return "00:00:00:00:00:%u%u" % (i + 1, i + 1)

    def host_ip_address(self, host_index, vlan_index):
        """Create a string of the host IP address"""
        if isinstance(vlan_index, (list, tuple)):
            vlan_index = vlan_index[0]
        return "10.%u.0.%u/%u" % (vlan_index + 1, host_index + 1, self.NETPREFIX)

    def host_ping(self, src_host, dst_ip, intf=None):
        """Default method to ping from one host to an IP address"""
        self.one_ipv4_ping(
            src_host,
            dst_ip,
            require_host_learned=False,
            retries=5,
            timeout=1000,
            intf=intf,
        )

    def set_host_ip(self, host, host_ip):
        """Default method for setting a hosts IP address"""
        host.setIP(str(host_ip.ip), prefixLen=self.NETPREFIX)

    def build_net(
        self,
        host_links=None,
        host_vlans=None,
        switch_links=None,
        link_vlans=None,
        mininet_host_options=None,
        n_vlans=1,
        dp_options=None,
        host_options=None,
        link_options=None,
        vlan_options=None,
        routers=None,
        router_options=None,
        include=None,
        include_optional=None,
    ):
        """
        Args:
            host_links (dict): Host index key to list of dp indices
            host_vlans (dict): Host index key to vlan index/indices
            switch_links (list): List of link tuples of switch indices (u, v)
            link_vlans (dict): Link tuple of switch indices (u, v) mapping to vlans
            mininet_host_options (dict): Host index map to additional mininet host options
            n_vlans (int): Number of VLANs to generate
            dp_options (dict): Additional options for each DP, keyed by DP index
            host_options (dict): Additional options for each host, keyed by host index
            link_options (dict): Additional options for each link, keyed by indices tuple (u, v)
            vlan_options (dict): Additional options for each VLAN, keyed by vlan index
            routers (dict): Router index to list of VLANs in the router
            router_options (dict): Additional options for each router, keyed by router index
            include (list): Files to include using the the Faucet config 'include' key
            include_optional (list): File to include using the Faucet config 'include_optional' key
        """
        self.topo = FaucetTopoGenerator(
            self.OVS_TYPE,
            self.ports_sock,
            self._test_name(),
            self.NUM_DPS,
            self.DESCENDING_DPIDS,
            host_links=host_links,
            host_vlans=host_vlans,
            switch_links=switch_links,
            link_vlans=link_vlans,
            hw_dpid=self.hw_dpid,
            hw_ports=self.switch_map,
            port_order=self.port_order,
            start_port=self.start_port,
            host_options=mininet_host_options,
        )
        self.dpids = self.topo.get_dpids()
        self.dpid = self.dpids[0]
        # host_port_maps = {host_n: {switch_n: [ports, ...], ...}, ...}
        # link_port_maps = {(switch_n, switch_m): [ports, ...], ...}
        (
            port_maps,
            self.host_port_maps,
            self.link_port_maps,
        ) = self.topo.create_port_maps()
        self.port_map = port_maps[self.dpid]
        dpid_names = {}
        # pylint: disable=consider-using-dict-items
        for i in self.topo.switches_by_id:
            dpid = self.topo.dpids_by_id[i]
            name = self.topo.switches_by_id[i]
            dpid_names[dpid] = name
        self.set_dpid_names(dpid_names)
        self.configuration_options = {
            "acl_options": self.acls(),
            "dp_options": dp_options,
            "host_options": host_options,
            "link_options": link_options,
            "vlan_options": vlan_options,
            "routers": routers,
            "router_options": router_options,
            "include": include,
            "include_optional": include_optional,
            "ignored_switches": self.IGNORED_SWITCHES,
        }
        self.CONFIG = self.topo.get_config(
            n_vlans,
            **self.configuration_options,
        )
        self.mininet_host_options = mininet_host_options
        self.n_vlans = n_vlans
        self.host_vlans = host_vlans
        self.link_vlans = link_vlans

    def start_net(self):
        """
        Override start_net to create the faucet vips, the host information and set up the
            host routes for routed hosts
        """
        super().start_net()
        # Create a dictionary of host information
        self.host_information = {}
        ips_for_vlans = {}
        for host_id, host_name in self.topo.hosts_by_id.items():
            host = self.net.get(host_name)
            vlan = self.host_vlans[host_id]
            ip_interface = None
            if vlan is not None:
                if isinstance(vlan, list):
                    vlan = tuple(vlan)
                ips_for_vlans.setdefault(vlan, 0)
                ip_addr = self.host_ip_address(ips_for_vlans[vlan], vlan)
                ips_for_vlans[vlan] += 1
                if self.mininet_host_options and host_id in self.mininet_host_options:
                    mininet_ip = self.mininet_host_options[host_id].get("ip", None)
                    if mininet_ip:
                        ip_addr = mininet_ip
                ip_interface = ipaddress.ip_interface(ip_addr)
                self.set_host_ip(host, ip_interface)
            self.host_information[host_id] = {
                "host": host,
                "ip": ip_interface,
                "mac": host.MAC(),
                "vlan": vlan,
                "bond": None,
                "ports": self.host_port_maps[host_id],
            }
        # Store faucet vip interfaces
        self.faucet_vips = {}
        for vlan in range(self.n_vlans):
            self.faucet_vips[vlan] = ipaddress.ip_interface(self.faucet_vip(vlan))
        # Setup the linux bonds for LACP connected hosts
        self.setup_lacp_bonds()
        # Add host routes to hosts for inter vlan routing
        self.setup_intervlan_host_routes()

    def setup_lacp_bonds(self):
        """Search through host options for lacp hosts and configure accordingly"""
        host_options = self.configuration_options["host_options"]
        if not host_options:
            return
        bond_index = 1
        for host_id, options in host_options.items():
            if "lacp" in options:
                host = self.host_information[host_id]["host"]
                # LACP must be configured with host ports down
                for dp_i, ports in self.host_port_maps[host_id].items():
                    for port in ports:
                        self.set_port_down(port, self.topo.dpids_by_id[dp_i])
                orig_ip = host.IP()
                lacp_switches = [
                    self.net.get(self.topo.switches_by_id[i])
                    for i in self.host_port_maps[host_id]
                ]
                bond_members = [
                    pair[0].name
                    for switch in lacp_switches
                    for pair in host.connectionsTo(switch)
                ]
                bond_name = "bond%u" % (bond_index)
                self.host_information[host_id]["bond"] = bond_name
                bond_cmds = []
                for bond_member in bond_members:
                    # Deconfigure bond members
                    bond_cmds.extend(
                        [
                            "link set %s down" % bond_member,
                            "address flush dev %s" % bond_member,
                        ]
                    )
                # Configure bond interface
                bond_cmds.extend(
                    [
                        (
                            "link add %s address 0e:00:00:00:00:99 "
                            "type bond mode 802.3ad lacp_rate fast miimon 100 "
                            "xmit_hash_policy layer2+3"
                        )
                        % (bond_name),
                        "addr add %s/%s dev %s" % (orig_ip, self.NETPREFIX, bond_name),
                        "link set %s up" % bond_name,
                    ]
                )
                # Add bond members
                for bond_member in bond_members:
                    bond_cmds.extend(
                        [
                            "link set dev %s master %s" % (bond_member, bond_name),
                        ]
                    )
                host.run_ip_batch(bond_cmds)
                bond_index += 1
                # Return the ports to UP
                for dp_i, ports in self.host_port_maps[host_id].items():
                    for port in ports:
                        self.set_port_up(port, self.topo.dpids_by_id[dp_i])

    def setup_intervlan_host_routes(self):
        """Configure host routes between hosts that belong on routed VLANs"""
        if self.configuration_options["routers"]:
            for src_name, src in self.host_information.items():
                src_host = src["host"]
                src_vlan = src["vlan"]
                src_ip = src["ip"]
                for dst_name, dst in self.host_information.items():
                    if src_name != dst_name:
                        dst_host = dst["host"]
                        dst_vlan = dst["vlan"]
                        dst_ip = dst["ip"]
                        if src_vlan != dst_vlan and self.is_routed_vlans(
                            src_vlan, dst_vlan
                        ):
                            src_faucet_vip = self.faucet_vips[src_vlan]
                            dst_faucet_vip = self.faucet_vips[dst_vlan]
                            self.add_host_route(src_host, dst_ip, src_faucet_vip.ip)
                            self.add_host_route(dst_host, src_ip, dst_faucet_vip.ip)

    def debug(self):
        """Print additional information when debugging"""
        try:
            super().debug()
        except Exception:
            pprint.pprint(self.host_information)
            raise

    def verify_no_cable_errors(self):
        """Check that prometheus does not detect any stack cabling errors on all DPs"""
        for i, name in self.topo.switches_by_id.items():
            dpid = self.dpids[i]
            labels = {"dp_id": "0x%x" % int(dpid), "dp_name": name}
            self.assertEqual(
                0,
                self.scrape_prometheus_var(
                    var="stack_cabling_errors_total", labels=labels, default=None
                ),
            )
            self.assertGreater(
                self.scrape_prometheus_var(
                    var="stack_probes_received_total", labels=labels
                ),
                0,
            )

    def verify_stack_hosts(self, verify_bridge_local_rule=True, retries=3):
        """Verify hosts with stack LLDP messages"""
        lldp_cap_files = []
        for host in self.hosts_name_ordered():
            lldp_cap_file = os.path.join(self.tmpdir, "%s-lldp.cap" % host)
            lldp_cap_files.append(lldp_cap_file)
            host.cmd(
                timeout_cmd(
                    "tcpdump -U -n -c 1 -i %s -w %s ether proto 0x88CC and not ether src %s &"
                    % (host.defaultIntf(), host.MAC(), lldp_cap_file),
                    60,
                )
            )
        # should not flood LLDP from hosts
        self.verify_lldp_blocked(self.hosts_name_ordered())
        # hosts should see no LLDP probes
        self.verify_empty_caps(lldp_cap_files)
        if verify_bridge_local_rule:
            # Verify 802.1x flood block triggered.
            for dpid in self.dpids:
                self.wait_nonzero_packet_count_flow(
                    {"dl_dst": "01:80:c2:00:00:00/ff:ff:ff:ff:ff:f0"},
                    dpid=dpid,
                    table_id=self._FLOOD_TABLE,
                    ofa_match=False,
                )
        self.retry_net_ping(retries=retries)

    def stack_port_status(self, dpid, dp_name, port_no):
        """Return the status of a stack port from prometheus"""
        labels = self.port_labels(port_no)
        labels.update({"dp_id": "0x%x" % int(dpid), "dp_name": dp_name})
        return self.scrape_prometheus_var(
            "port_stack_state", labels=labels, default=None, dpid=dpid
        )

    def wait_for_stack_port_status(self, dpid, dp_name, port_no, status, timeout=25):
        """Wait until prometheus detects a stack port has a certain status"""
        labels = self.port_labels(port_no)
        labels.update({"dp_id": "0x%x" % int(dpid), "dp_name": dp_name})
        if not self.wait_for_prometheus_var(
            "port_stack_state",
            status,
            labels=labels,
            default=None,
            dpid=False,
            timeout=timeout,
        ):
            self.fail(
                "did not get expected dpid %x port %u port_stack_state %u"
                % (int(dpid), port_no, status)
            )

    def one_stack_port_down(self, dpid, dp_name, port):
        """Set a stack port down and wait for prometheus to detect the change"""
        self.set_port_down(port, dpid, wait=False)
        self.wait_for_stack_port_status(dpid, dp_name, port, 4)

    def one_stack_port_up(self, dpid, dp_name, port):
        """Set a stack port up and wait for prometheus to detect the change"""
        self.set_port_up(port, dpid, wait=False)
        self.wait_for_stack_port_status(dpid, dp_name, port, 3)

    def verify_stack_up(self, prop=1.0, timeout=25):
        """Verify all stack ports are up"""
        for _ in range(timeout):
            links = 0
            links_up = 0
            for link, ports in self.link_port_maps.items():
                if link[0] in self.IGNORED_SWITCHES or link[1] in self.IGNORED_SWITCHES:
                    continue
                dpid = self.topo.dpids_by_id[link[0]]
                name = self.topo.switches_by_id[link[0]]
                for port in ports:
                    status = self.stack_port_status(dpid, name, port)
                    links += 1
                    if status == 3:  # STACK_STATE_UP
                        links_up += 1
            prop_up = links_up / links
            if prop_up >= prop:
                return
            time.sleep(1)
        self.fail("not enough links up: %f / %f" % (links_up, links))

    def verify_one_stack_down(self, stack_offset_port, coldstart=False):
        """Test conditions when one stack port is down"""
        self.retry_net_ping()
        stack_link = None
        count = 0
        for sport, link in self.topo.ports[self.topo.switches_by_id[0]].items():
            if self.topo.isSwitch(link[0]):
                if count == stack_offset_port:
                    stack_link = (sport, link[1])
                    break
                count += 1
        stack_port, remote_stack_port = stack_link  # pytype: disable=attribute-error
        self.set_port_down(stack_port, wait=False)
        # self.dpids[1] is the intermediate switch.
        self.set_port_down(remote_stack_port, self.dpids[1], wait=False)
        # test case where one link is down when coldstarted.
        if coldstart:
            self.coldstart_conf()
        self.verify_stack_up(prop=0.75)
        self.verify_stack_hosts(verify_bridge_local_rule=False)
        # Broadcast works, and first switch doesn't see broadcast packet ins from stack.
        packet_in_before_broadcast = self.scrape_prometheus_var("of_vlan_packet_ins")
        self.verify_broadcast()
        packet_in_after_broadcast = self.scrape_prometheus_var("of_vlan_packet_ins")
        self.assertEqual(packet_in_before_broadcast, packet_in_after_broadcast)
        self.verify_no_cable_errors()

    def verify_no_arp_storm(self, ping_host, tcpdump_host):
        """Check that there is no excess ARP packets in the network"""
        switch_to_switch_links = 0
        for link in self.topo.links():
            src_node, dst_node = link
            if self.topo.isSwitch(src_node):
                if self.topo.isSwitch(dst_node):
                    switch_to_switch_links += 1
        num_arp_expected = switch_to_switch_links * 2
        tcpdump_filter = "arp and ether src %s" % ping_host.MAC()
        tcpdump_txt = self.tcpdump_helper(
            tcpdump_host,
            tcpdump_filter,
            [
                lambda: ping_host.cmd("arp -d %s" % tcpdump_host.IP()),
                lambda: ping_host.cmd("ping -c1 %s" % tcpdump_host.IP()),
            ],
            packets=(num_arp_expected + 1),
        )
        num_arp_received = len(
            re.findall(
                "who-has %s tell %s" % (tcpdump_host.IP(), ping_host.IP()), tcpdump_txt
            )
        )
        self.assertTrue(num_arp_received)
        self.assertLessEqual(num_arp_received, num_arp_expected)

    def verify_stack_has_no_loop(self):
        """Ping between first and last hosts (by name) then verify there is no broadcast storm"""
        for ping_host, tcpdump_host in (
            (self.hosts_name_ordered()[0], self.hosts_name_ordered()[-1]),
            (self.hosts_name_ordered()[-1], self.hosts_name_ordered()[0]),
        ):
            self.verify_no_arp_storm(ping_host, tcpdump_host)

    def verify_all_stack_hosts(self):
        """Test conditions for stack hosts"""
        for _ in range(2):
            self.verify_stack_up()
            self.verify_no_cable_errors()
            self.verify_stack_hosts()
            self.verify_traveling_dhcp_mac()
            self.verify_unicast_not_looped()
            self.verify_no_bcast_to_self()
            self.verify_stack_has_no_loop()
            self.flap_all_switch_ports()

    def verify_tunnel_established(
        self, src_host, dst_host, other_host, packets=3, dpid=None
    ):
        """Verify ICMP packets tunnelled from src to dst."""
        icmp_match = {"eth_type": IPV4_ETH, "ip_proto": 1}
        self.wait_until_matching_flow(
            icmp_match, table_id=self._PORT_ACL_TABLE, ofa_match=False, dpid=dpid
        )
        tcpdump_text = self.tcpdump_helper(
            dst_host,
            "icmp[icmptype] == 8",
            [
                # need to set static ARP as only ICMP is tunnelled.
                lambda: src_host.cmd(
                    "arp -s %s %s" % (other_host.IP(), other_host.MAC())
                ),
                lambda: src_host.cmd("ping -c%u -t1 %s" % (packets, other_host.IP())),
            ],
            packets=1,
            timeout=(packets + 1),
        )
        self.wait_nonzero_packet_count_flow(
            icmp_match, table_id=self._PORT_ACL_TABLE, ofa_match=False, dpid=dpid
        )
        self.assertTrue(
            re.search("%s: ICMP echo request" % other_host.IP(), tcpdump_text),
            "Tunnel was not established",
        )

    def verify_one_broadcast(self, from_host, to_hosts):
        """Verify host connectivity via broadcast"""
        self.assertGreater(len(to_hosts), 1, "Testing only one ext host is not useful")
        received_broadcasts = []
        for to_host in to_hosts:
            if self.verify_broadcast(
                hosts=(from_host, to_host), broadcast_expected=None
            ):
                received_broadcasts.append(to_host)
        received_names = {host.name: host for host in received_broadcasts}
        self.assertEqual(
            len(received_broadcasts),
            1,
            "Received not exactly one broadcast from %s: %s"
            % (from_host.name, received_names),
        )

    def map_int_ext_hosts(self):
        """
        Obtains a list of the interal hosts, the external hosts and a dictionary
            of the internal and external hosts for each DP by DP name
        Returns int_hosts, ext_hosts, dp_hosts
        """
        int_hosts = []
        ext_hosts = []
        dp_hosts = {
            self.topo.switches_by_id[dp_index]: ([], [])
            for dp_index in range(self.NUM_DPS)
        }
        for host_id, options in self.configuration_options["host_options"].items():
            host = self.host_information[host_id]["host"]
            if options.get("loop_protect_external", False):
                ext_hosts.append(host)
                int_or_ext = 1
            else:
                int_hosts.append(host)
                int_or_ext = 0
            for dp_i in self.host_port_maps[host_id].keys():
                switch = self.topo.switches_by_id[dp_i]
                if isinstance(dp_hosts[switch][int_or_ext], list):
                    dp_hosts[switch][int_or_ext].append(host)
        return set(int_hosts), set(ext_hosts), dp_hosts

    def verify_protected_connectivity(self):
        """
        Checks:
            - All internal hosts can reach other internal hosts
            - All internal hosts can reach exactly one external host
            - All external hosts cannot flood to each other
            - All external hosts can reach the internal hosts
        """
        self.verify_stack_up()
        int_hosts, ext_hosts, _ = self.map_int_ext_hosts()

        for int_host in int_hosts:
            # All internal hosts can reach other internal hosts.
            for other_int_host in int_hosts - {int_host}:
                self.verify_broadcast(
                    hosts=(int_host, other_int_host), broadcast_expected=True
                )
                self.one_ipv4_ping(int_host, other_int_host.IP())

            # All internal hosts can reach exactly one external host.
            self.verify_one_broadcast(int_host, ext_hosts)

        for ext_host in ext_hosts:
            # All external hosts can reach internal hosts.
            for int_host in int_hosts:
                self.verify_broadcast(
                    hosts=(ext_host, int_host), broadcast_expected=True
                )
                self.one_ipv4_ping(ext_host, int_host.IP())

            # All external hosts cannot flood to each other.
            for other_ext_host in ext_hosts - {ext_host}:
                self.verify_broadcast(
                    hosts=(ext_host, other_ext_host), broadcast_expected=False
                )

    def set_externals_state(self, dp_name, externals_up):
        """Set the port up/down state of all external ports on a switch"""
        dp_conf = self._get_faucet_conf()["dps"][dp_name]
        for port_num, port_conf in dp_conf["interfaces"].items():
            if port_conf.get("loop_protect_external"):
                if externals_up:
                    self.set_port_up(port_num, dp_conf.get("dp_id"))
                else:
                    self.set_port_down(port_num, dp_conf.get("dp_id"))

    def validate_with_externals_down(self, dp_name):
        """Check situation when all externals on a given dp are down"""
        self.set_externals_state(dp_name, False)
        self.verify_protected_connectivity()
        self.set_externals_state(dp_name, True)

    def validate_with_externals_down_fails(self, dp_name):
        """Faucet code is not currently correct, so expect to fail."""
        # TODO: Fix faucet so the test inversion is no longer required.
        asserted = False
        try:
            self.validate_with_externals_down(dp_name)
        except AssertionError:
            asserted = True
        self.assertTrue(asserted, "Did not fail as expected for %s" % dp_name)

    def verify_intervlan_routing(self):
        """Verify intervlan routing but for LAG host use bond interface"""
        for src in self.host_information:
            for dst in self.host_information:
                if dst > src:
                    self.check_host_connectivity_by_id(src, dst)

    def check_host_connectivity_by_id(self, src_id, dst_id):
        """Ping from src to dst with host_id parameters if they should be able to"""
        src_host, src_ip, _, src_vlan, src_bond, _ = self.host_information[
            src_id
        ].values()
        dst_host, dst_ip, _, dst_vlan, dst_bond, _ = self.host_information[
            dst_id
        ].values()
        connectivity = src_vlan == dst_vlan or self.is_routed_vlans(src_vlan, dst_vlan)
        if self.is_routed_vlans(src_vlan, dst_vlan):
            src_vip = self.faucet_vips[src_vlan]
            dst_vip = self.faucet_vips[dst_vlan]
            # pytype: disable=attribute-error
            self.host_ping(src_host, src_vip.ip, src_bond)
            self.host_ping(dst_host, dst_vip.ip, dst_bond)
        if connectivity:
            # pytype: disable=attribute-error
            self.host_ping(src_host, dst_ip.ip, src_bond)
            self.host_ping(dst_host, src_ip.ip, dst_bond)

    def is_routed_vlans(self, vlan_a, vlan_b):
        """Return true if the two vlans share a router"""
        if self.configuration_options["routers"]:
            for vlans in self.configuration_options["routers"].values():
                if vlan_a in vlans and vlan_b in vlans:
                    return True
        return False

    def bcast_dst_blocked_helper(
        self, port, first_host, second_host, success_re, retries
    ):
        """Helper for checking broadcast destination has been blocked"""
        tcpdump_filter = "udp and ether src %s and ether dst %s" % (
            first_host.MAC(),
            "ff:ff:ff:ff:ff:ff",
        )
        target_addr = str(self.faucet_vips[0].network.broadcast_address)
        for _ in range(retries):
            tcpdump_txt = self.tcpdump_helper(
                second_host,
                tcpdump_filter,
                [
                    partial(
                        first_host.cmd,
                        (
                            "date | socat - udp-datagram:%s:%d,broadcast"
                            % (target_addr, port)
                        ),
                    )
                ],
                packets=1,
            )
            if re.search(success_re, tcpdump_txt):
                return True
            time.sleep(1)
        return False

    def get_expected_synced_states(self, host_id):
        """Return the list of regex string for the expected sync state of a LACP LAG connection"""
        synced_state_list = []
        oper_key = self.configuration_options["host_options"][host_id]["lacp"]
        lacp_ports = [
            port
            for ports in self.host_information[host_id]["ports"].values()
            for port in ports
        ]
        for port in lacp_ports:
            synced_state_txt = r"""
Slave Interface: \S+
MII Status: up
Speed: \d+ Mbps
Duplex: full
Link Failure Count: \d+
Permanent HW addr: \S+
Slave queue ID: 0
Aggregator ID: \d+
Actor Churn State: monitoring
Partner Churn State: monitoring
Actor Churned Count: \d+
Partner Churned Count: \d+
details actor lacp pdu:
    system priority: 65535
    system mac address: 0e:00:00:00:00:99
    port key: \d+
    port priority: 255
    port number: \d+
    port state: 63
details partner lacp pdu:
    system priority: 65535
    system mac address: 0e:00:00:00:00:01
    oper key: %d
    port priority: 255
    port number: %d
    port state: 62
""".strip() % (
                oper_key,
                port,
            )
            synced_state_list.append(synced_state_txt)
        return synced_state_list

    def prom_lacp_up_ports(self, dpid):
        """Get the number of up LAG ports according to Prometheus for a dpid"""
        lacp_up_ports = 0
        for host_id, options in self.configuration_options["host_options"].items():
            # Find LACP hosts
            for key in options.keys():
                if key == "lacp":
                    # Is LACP host
                    for dp_i, ports in self.host_port_maps[host_id].items():
                        if dpid == self.topo.dpids_by_id[dp_i]:
                            # Host has links to dpid
                            for port in ports:
                                # Obtain up LACP ports for that dpid
                                port_labels = self.port_labels(port)
                                lacp_state = self.scrape_prometheus_var(
                                    "port_lacp_state", port_labels, default=0, dpid=dpid
                                )
                                lacp_up_ports += 1 if lacp_state == 3 else 0
        return lacp_up_ports

    def verify_num_lag_up_ports(self, expected_up_ports, dpid):
        """Checks to see if Prometheus has the expected number of up LAG ports
        on the specified DP"""
        for _ in range(self.LACP_TIMEOUT * 10):
            if self.prom_lacp_up_ports(dpid) == expected_up_ports:
                return
            time.sleep(1)
        self.assertEqual(self.prom_lacp_up_ports(dpid), expected_up_ports)

    def require_linux_bond_up(self, host_id):
        """Checks to see if the host has properly formed into a bonded state"""
        synced_state_list = self.get_expected_synced_states(host_id)
        host = self.host_information[host_id]["host"]
        bond_name = self.host_information[host_id]["bond"]
        for _ in range(self.LACP_TIMEOUT * 2):
            result = host.cmd('cat /proc/net/bonding/%s|sed "s/[ \t]*$//g"' % bond_name)
            result = "\n".join([line.rstrip() for line in result.splitlines()])
            with open(
                os.path.join(self.tmpdir, "bonding-state.txt"), "w", encoding="utf-8"
            ) as state_file:
                state_file.write(result)
            matched_all = True
            for state_txt in synced_state_list:
                if not re.search(state_txt, result):
                    matched_all = False
                    break
            if matched_all:
                return
            time.sleep(1)
        synced_state_txt = r""""""
        for state_txt in synced_state_list:
            synced_state_txt += state_txt + "\n\n"
        synced_state_txt.strip()
        self.assertFalse(
            re.search(synced_state_txt, result),
            msg="LACP did not synchronize: %s\n\nexpected:\n\n%s"
            % (result, synced_state_txt),
        )

    def verify_lag_connectivity(self, host_id):
        """Verify LAG connectivity"""
        lacp_ports = self.host_port_maps[host_id]
        # All ports down
        for dp_i, ports in lacp_ports.items():
            dpid = self.topo.dpids_by_id[dp_i]
            for port in ports:
                self.set_port_down(port, dpid)
            self.verify_num_lag_up_ports(0, dpid)
        # Pick a port to set up
        up_dp = random.choice(list(lacp_ports.keys()))
        up_dpid = self.topo.dpids_by_id[up_dp]
        up_port = random.choice(lacp_ports[up_dp])
        self.set_port_up(up_port, up_dpid)
        self.verify_num_lag_up_ports(1, up_dpid)
        # Ensure connectivity with one port
        self.verify_lag_host_connectivity()
        # Set the other ports to UP
        for dp_i, ports in lacp_ports.items():
            dpid = self.topo.dpids_by_id[dp_i]
            for port in ports:
                self.set_port_up(port, dpid)
            self.verify_num_lag_up_ports(len(ports), dpid)
        # Ensure connectivity with all ports
        self.require_linux_bond_up(host_id)
        self.verify_lag_host_connectivity()
        # Tear down first port
        self.set_port_down(up_port, up_dpid)
        self.verify_num_lag_up_ports(len(lacp_ports[up_dp]) - 1, up_dpid)
        # Ensure connectivity with new ports only
        self.verify_lag_host_connectivity()

    def verify_lag_host_connectivity(self):
        """Verify LAG hosts can connect to any other host using the interface"""
        # Find all LACP hosts
        for lacp_id, host_options in self.configuration_options["host_options"].items():
            if "lacp" in host_options:
                # Found LACP host
                for dst_id in self.host_information:
                    if lacp_id == dst_id:
                        continue
                    # Test connectivity to any other host (might be another LAG host)
                    self.check_host_connectivity_by_id(lacp_id, dst_id)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2012 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-class-docstring,disable=missing-function-docstring,disable=invalid-name
# pylint: disable=import-outside-toplevel,disable=too-few-public-methods

import inspect
from types import MethodType

from routes import Mapper
from routes.util import URLGenerator
import six
from tinyrpc.server import RPCServer
from tinyrpc.dispatch import RPCDispatcher
from tinyrpc.protocols.jsonrpc import JSONRPCProtocol
from tinyrpc.transports import ServerTransport, ClientTransport
from tinyrpc.client import RPCClient
import webob.dec
import webob.exc
from webob.request import Request as webob_Request
from webob.response import Response as webob_Response

from os_ken.lib import hub

HEX_PATTERN = r"0x[0-9a-z]+"
DIGIT_PATTERN = r"[1-9][0-9]*"


def route(name, path, methods=None, requirements=None):
    def _route(controller_method):
        controller_method.routing_info = {
            "name": name,
            "path": path,
            "methods": methods,
            "requirements": requirements,
        }
        return controller_method

    return _route


class Request(webob_Request):
    """
    Wrapper class for webob.request.Request.

    The behavior of this class is the same as webob.request.Request
    except for setting "charset" to "UTF-8" automatically.
    """

    DEFAULT_CHARSET = "UTF-8"

    def __init__(self, environ, *args, charset=DEFAULT_CHARSET, **kwargs):
        super().__init__(environ, *args, **kwargs, charset=charset)


class Response(webob_Response):
    """
    Wrapper class for webob.response.Response.

    The behavior of this class is the same as webob.response.Response
    except for setting "charset" to "UTF-8" automatically.
    """

    DEFAULT_CHARSET = "UTF-8"

    def __init__(self, *args, charset=DEFAULT_CHARSET, **kwargs):
        super().__init__(*args, **kwargs, charset=charset)


class WebSocketRegistrationWrapper:
    def __init__(self, func, controller):
        self._controller = controller
        self._controller_method = MethodType(func, controller)

    def __call__(self, ws):
        wsgi_application = self._controller.parent
        ws_manager = wsgi_application.websocketmanager
        ws_manager.add_connection(ws)
        try:
            self._controller_method(ws)
        finally:
            ws_manager.delete_connection(ws)


class _AlreadyHandledResponse(Response):
    # XXX: Eventlet API should not be used directly.
    # https://github.com/benoitc/gunicorn/pull/2581
    from packaging import version
    import eventlet

    if version.parse(eventlet.__version__) >= version.parse("0.30.3"):
        import eventlet.wsgi

        _ALREADY_HANDLED = getattr(eventlet.wsgi, "ALREADY_HANDLED", None)
    else:
        from eventlet.wsgi import ALREADY_HANDLED  # pylint: disable=no-name-in-module

        _ALREADY_HANDLED = ALREADY_HANDLED

    def __call__(self, environ, start_response):
        return self._ALREADY_HANDLED


def websocket(name, path):
    def _websocket(controller_func):
        def __websocket(self, req, **_):
            wrapper = WebSocketRegistrationWrapper(controller_func, self)
            ws_wsgi = hub.WebSocketWSGI(wrapper)
            ws_wsgi(req.environ, req.start_response)
            # XXX: In order to prevent the writing to a already closed socket.
            #      This issue is caused by combined use:
            #       - webob.dec.wsgify()
            #       - eventlet.wsgi.HttpProtocol.handle_one_response()
            return _AlreadyHandledResponse()

        __websocket.routing_info = {
            "name": name,
            "path": path,
            "methods": None,
            "requirements": None,
        }
        return __websocket

    return _websocket


class ControllerBase:
    special_vars = ["action", "controller"]

    def __init__(self, req, link, data, **config):
        self.req = req
        self.link = link
        self.data = data
        self.parent = None
        for name, value in config.items():
            setattr(self, name, value)

    def __call__(self, req):
        action = self.req.urlvars.get("action", "index")
        if hasattr(self, "__before__"):
            self.__before__()

        kwargs = self.req.urlvars.copy()
        for attr in self.special_vars:
            if attr in kwargs:
                del kwargs[attr]

        return getattr(self, action)(req, **kwargs)


class WebSocketDisconnectedError(Exception):
    pass


class WebSocketServerTransport(ServerTransport):
    def __init__(self, ws):
        self.ws = ws

    def receive_message(self):
        message = self.ws.wait()
        if message is None:
            raise WebSocketDisconnectedError()
        context = None
        return context, message

    def send_reply(self, context, reply):
        self.ws.send(six.text_type(reply))


class WebSocketRPCServer(RPCServer):
    def __init__(self, ws, rpc_callback):
        dispatcher = RPCDispatcher()
        dispatcher.register_instance(rpc_callback)
        super().__init__(
            WebSocketServerTransport(ws),
            JSONRPCProtocol(),
            dispatcher,
        )

    def serve_forever(self):
        try:
            super().serve_forever()
        except WebSocketDisconnectedError:
            return

    def _spawn(self, func, *args, **kwargs):
        hub.spawn(func, *args, **kwargs)


class WebSocketClientTransport(ClientTransport):
    def __init__(self, ws, queue):
        self.ws = ws
        self.queue = queue

    def send_message(self, message, expect_reply=True):
        self.ws.send(six.text_type(message))

        if expect_reply:
            return self.queue.get()
        return None


class WebSocketRPCClient(RPCClient):
    def __init__(self, ws):
        self.ws = ws
        self.queue = hub.Queue()
        super().__init__(
            JSONRPCProtocol(),
            WebSocketClientTransport(ws, self.queue),
        )

    def serve_forever(self):
        while True:
            msg = self.ws.wait()
            if msg is None:
                break
            self.queue.put(msg)


class wsgify_hack(webob.dec.wsgify):
    def __call__(self, environ, start_response):
        self.kwargs["start_response"] = start_response
        return super().__call__(environ, start_response)


class WebSocketManager:
    def __init__(self):
        self._connections = []

    def add_connection(self, ws):
        self._connections.append(ws)

    def delete_connection(self, ws):
        self._connections.remove(ws)

    def broadcast(self, msg):
        for connection in self._connections:
            connection.send(msg)


class WSGIApplication:
    def __init__(self, **config):
        self.config = config
        self.mapper = Mapper()
        self.registory = {}
        self._wsmanager = WebSocketManager()
        super().__init__()

    def _match(self, req):
        # Note: Invoke the new API, first. If the arguments unmatched,
        # invoke the old API.
        try:
            return self.mapper.match(environ=req.environ)
        except TypeError:
            self.mapper.environ = req.environ
            return self.mapper.match(req.path_info)

    @wsgify_hack
    def __call__(self, req, start_response):
        match = self._match(req)

        if not match:
            return webob.exc.HTTPNotFound()

        req.start_response = start_response
        req.urlvars = match
        link = URLGenerator(self.mapper, req.environ)

        data = None
        name = match["controller"].__name__
        if name in self.registory:
            data = self.registory[name]

        controller = match["controller"](req, link, data, **self.config)
        controller.parent = self
        return controller(req)

    def register(self, controller, data=None):
        def _target_filter(attr):
            if not inspect.ismethod(attr) and not inspect.isfunction(attr):
                return False
            if not hasattr(attr, "routing_info"):
                return False
            return True

        methods = inspect.getmembers(controller, _target_filter)
        for method_name, method in methods:
            routing_info = getattr(method, "routing_info")
            name = routing_info["name"]
            path = routing_info["path"]
            conditions = {}
            if routing_info.get("methods"):
                conditions["method"] = routing_info["methods"]
            requirements = routing_info.get("requirements") or {}
            self.mapper.connect(
                name,
                path,
                controller=controller,
                requirements=requirements,
                action=method_name,
                conditions=conditions,
            )
        if data:
            self.registory[controller.__name__] = data

    @property
    def websocketmanager(self):
        return self._wsmanager


class WSGIServer(hub.WSGIServer):
    def __init__(self, application, host, port, **config):
        super().__init__((host, port), application, **config)

    def __call__(self):
        self.serve_forever()



#!/usr/bin/env python3

"""Unit tests run as PYTHONPATH=../../.. python3 ./test_valve.py."""

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from os_ken.ofproto import ofproto_v1_3 as ofp

from clib.valve_test_lib import CONFIG, DP1_CONFIG, FAUCET_MAC, ValveTestBases


class ValveTestEgressPipeline(
    ValveTestBases.ValveTestBig
):  # pylint: disable=too-few-public-methods
    """Run complete set of basic tests."""

    DP1_CONFIG = (
        """
            egress_pipeline: True
    """
        + DP1_CONFIG
    )


class ValveEgressACLTestCase(ValveTestBases.ValveTestNetwork):
    """Test ACL drop/allow and reloading."""

    def setUp(self):
        self.setup_valves(CONFIG)

    def test_vlan_acl_deny(self):
        """Test VLAN ACL denies a packet."""
        ALLOW_HOST_V6 = "fc00:200::1:1"  # pylint: disable=invalid-name
        DENY_HOST_V6 = "fc00:200::1:2"  # pylint: disable=invalid-name
        FAUCET_V100_VIP = "fc00:100::1"  # pylint: disable=invalid-name
        FAUCET_V200_VIP = "fc00:200::1"  # pylint: disable=invalid-name
        acl_config = """
dps:
    s1:
{dp1_config}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
            p2:
                number: 2
                native_vlan: v200
                tagged_vlans: [v100]
            p3:
                number: 3
                tagged_vlans: [v100, v200]
            p4:
                number: 4
                tagged_vlans: [v200]
vlans:
    v100:
        vid: 0x100
        faucet_mac: '{mac}'
        faucet_vips: ['{v100_vip}/64']
    v200:
        vid: 0x200
        faucet_mac: '{mac}'
        faucet_vips: ['{v200_vip}/64']
        acl_out: drop_non_allow_host_v6
        minimum_ip_size_check: false
routers:
    r_v100_v200:
        vlans: [v100, v200]
acls:
    drop_non_allow_host_v6:
        - rule:
            ipv6_dst: '{allow_host}'
            eth_type: 0x86DD
            actions:
                allow: 1
        - rule:
            eth_type: 0x86DD
            actions:
                allow: 0
""".format(
            dp1_config=DP1_CONFIG,
            mac=FAUCET_MAC,
            v100_vip=FAUCET_V100_VIP,
            v200_vip=FAUCET_V200_VIP,
            allow_host=ALLOW_HOST_V6,
        )

        l2_drop_match = {
            "in_port": 2,
            "eth_dst": self.P3_V200_MAC,
            "vlan_vid": 0,
            "eth_type": 0x86DD,
            "ipv6_dst": DENY_HOST_V6,
        }
        l2_accept_match = {
            "in_port": 3,
            "eth_dst": self.P2_V200_MAC,
            "vlan_vid": 0x200 | ofp.OFPVID_PRESENT,
            "eth_type": 0x86DD,
            "ipv6_dst": ALLOW_HOST_V6,
        }
        v100_accept_match = {"in_port": 1, "vlan_vid": 0}
        table = self.network.tables[self.DP_ID]

        # base case
        for match in (l2_drop_match, l2_accept_match):
            self.assertTrue(
                table.is_output(match, port=4),
                msg="Packet not output before adding ACL",
            )

        # multicast
        self.update_config(acl_config, reload_type="cold")
        self.assertTrue(
            table.is_output(v100_accept_match, port=3),
            msg="Packet not output when on vlan with no ACL",
        )
        self.assertFalse(
            table.is_output(l2_drop_match, port=3), msg="Packet not blocked by ACL"
        )
        self.assertTrue(
            table.is_output(l2_accept_match, port=2), msg="Packet not allowed by ACL"
        )

        # unicast
        self.rcv_packet(
            2,
            0x200,
            {
                "eth_src": self.P2_V200_MAC,
                "eth_dst": self.P3_V200_MAC,
                "vid": 0x200,
                "ipv6_src": ALLOW_HOST_V6,
                "ipv6_dst": DENY_HOST_V6,
                "neighbor_advert_ip": ALLOW_HOST_V6,
            },
        )
        self.rcv_packet(
            3,
            0x200,
            {
                "eth_src": self.P3_V200_MAC,
                "eth_dst": self.P2_V200_MAC,
                "vid": 0x200,
                "ipv6_src": DENY_HOST_V6,
                "ipv6_dst": ALLOW_HOST_V6,
                "neighbor_advert_ip": DENY_HOST_V6,
            },
        )

        self.assertTrue(
            table.is_output(l2_accept_match, port=2), msg="Packet not allowed by ACL"
        )
        self.assertFalse(
            table.is_output(l2_drop_match, port=3), msg="Packet not blocked by ACL"
        )

        # l3
        l3_drop_match = {
            "in_port": 1,
            "eth_dst": FAUCET_MAC,
            "vlan_vid": 0,
            "eth_type": 0x86DD,
            "ipv6_dst": DENY_HOST_V6,
        }
        l3_accept_match = {
            "in_port": 1,
            "eth_dst": FAUCET_MAC,
            "vlan_vid": 0,
            "eth_type": 0x86DD,
            "ipv6_dst": ALLOW_HOST_V6,
        }

        self.assertTrue(
            table.is_output(l3_accept_match, port=2),
            msg="Routed packet not allowed by ACL",
        )
        self.assertFalse(
            table.is_output(l3_drop_match, port=3),
            msg="Routed packet not blocked by ACL",
        )


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr
